/** M6502: portable 6502 emulator ****************************/
/**                                                         **/
/**                         M6502.c                         **/
/**                                                         **/
/** This file contains implementation for 6502 CPU. Don't   **/
/** forget to provide nes_read_byte(), nes_write_byte(), Loop6502(), and   **/
/** possibly Op6502() functions to accomodate the emulated  **/
/** machine's architecture.                                 **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1996                      **/
/**               Alex Krasivsky  1996                      **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/

#include "M6502.h"
#include "Tables.h"
#include "emulate.h"

#ifndef TARGET_DC
#include <stdio.h>
#endif
/** INLINE ***************************************************/
/** Different compilers inline C functions differently.     **/
/*************************************************************/
#ifdef __GNUC__
#define INLINE inline
#else
#define INLINE static
#endif

/** System-Dependent Stuff ***********************************/
/** This is system-dependent code put here to speed things  **/
/** up. It has to stay inlined to be fast.                  **/
/*************************************************************/
#ifdef INES
#define FAST_RDOP
extern byte *Page[];
INLINE byte Op6502(register word A)
{
	return (Page[A >> 13][A & 0x1FFF]);
}
#endif

/** FAST_RDOP ************************************************/
/** With this #define not present, nes_read_byte() should perform  **/
/** the functions of nes_read_byte().                              **/
/*************************************************************/
#ifndef FAST_RDOP
#define Op6502(A) nes_read_byte(A)
/*prg_banks[(A & 0x6000) >> 13][A & 0x1fff]*/
#endif

/** Addressing Methods ***************************************/
/** These macros calculate and return effective addresses.  **/
/*************************************************************/
#define MC_Ab(Rg)	M_LDWORD(Rg)
#define MC_Zp(Rg)	Rg.B.l=Op6502(R->PC.W++);Rg.B.h=0
#define MC_Zx(Rg)	Rg.B.l=Op6502(R->PC.W++)+R->X;Rg.B.h=0
#define MC_Zy(Rg)	Rg.B.l=Op6502(R->PC.W++)+R->Y;Rg.B.h=0
#define MC_Ax(Rg)	M_LDWORD(Rg);Rg.W+=R->X
#define MC_Ay(Rg)	M_LDWORD(Rg);Rg.W+=R->Y
#define MC_Ix(Rg)	K.B.l=Op6502(R->PC.W++)+R->X;K.B.h=0; \
			Rg.B.l=Op6502(K.W++);Rg.B.h=Op6502(K.W)
#define MC_Iy(Rg)	K.B.l=Op6502(R->PC.W++);K.B.h=0; \
			Rg.B.l=Op6502(K.W++);Rg.B.h=Op6502(K.W); \
			Rg.W+=R->Y

/** Reading From Memory **************************************/
/** These macros calculate address and read from it.        **/
/*************************************************************/
#define MR_Ab(Rg)	MC_Ab(J);Rg=nes_read_byte(J.W)
#define MR_Im(Rg)	Rg=Op6502(R->PC.W++)
#define	MR_Zp(Rg)	MC_Zp(J);Rg=nes_read_byte(J.W)
#define MR_Zx(Rg)	MC_Zx(J);Rg=nes_read_byte(J.W)
#define MR_Zy(Rg)	MC_Zy(J);Rg=nes_read_byte(J.W)
#define	MR_Ax(Rg)	MC_Ax(J);Rg=nes_read_byte(J.W)
#define MR_Ay(Rg)	MC_Ay(J);Rg=nes_read_byte(J.W)
#define MR_Ix(Rg)	MC_Ix(J);Rg=nes_read_byte(J.W)
#define MR_Iy(Rg)	MC_Iy(J);Rg=nes_read_byte(J.W)

/** Writing To Memory ****************************************/
/** These macros calculate address and write to it.         **/
/*************************************************************/
#define MW_Ab(Rg)	MC_Ab(J);nes_write_byte(J.W,Rg)
#define MW_Zp(Rg)	MC_Zp(J);nes_write_byte(J.W,Rg)
#define MW_Zx(Rg)	MC_Zx(J);nes_write_byte(J.W,Rg)
#define MW_Zy(Rg)	MC_Zy(J);nes_write_byte(J.W,Rg)
#define MW_Ax(Rg)	MC_Ax(J);nes_write_byte(J.W,Rg)
#define MW_Ay(Rg)	MC_Ay(J);nes_write_byte(J.W,Rg)
#define MW_Ix(Rg)	MC_Ix(J);nes_write_byte(J.W,Rg)
#define MW_Iy(Rg)	MC_Iy(J);nes_write_byte(J.W,Rg)

/** Modifying Memory *****************************************/
/** These macros calculate address and modify it.           **/
/*************************************************************/
#define MM_Ab(Cmd)	MC_Ab(J);I=nes_read_byte(J.W);Cmd(I);nes_write_byte(J.W,I)
#define MM_Zp(Cmd)	MC_Zp(J);I=nes_read_byte(J.W);Cmd(I);nes_write_byte(J.W,I)
#define MM_Zx(Cmd)	MC_Zx(J);I=nes_read_byte(J.W);Cmd(I);nes_write_byte(J.W,I)
#define MM_Ax(Cmd)	MC_Ax(J);I=nes_read_byte(J.W);Cmd(I);nes_write_byte(J.W,I)

/** Other Macros *********************************************/
/** Calculating flags, stack, jumps, arithmetics, etc.      **/
/*************************************************************/
#define M_FL(Rg)	R->P=(R->P&~(Z_FLAG|N_FLAG))|ZNTable[Rg]
#define M_LDWORD(Rg)	Rg.B.l=Op6502(R->PC.W++);Rg.B.h=Op6502(R->PC.W++)

#define M_PUSH(Rg)	nes_write_byte(0x0100|R->S,Rg);R->S--
#define M_POP(Rg)	R->S++;Rg=Op6502(0x0100|R->S)
#define M_JR		R->PC.W+=(offset)Op6502(R->PC.W)+1;R->ICount--

#define M_ADC(Rg) \
  if(R->P&D_FLAG) \
  { \
    K.B.l=(R->A&0x0F)+(Rg&0x0F)+(R->P&C_FLAG); \
    K.B.h=(R->A>>4)+(Rg>>4)+(K.B.l>15? 1:0); \
    if(K.B.l>9) { K.B.l+=6;K.B.h++; } \
    if(K.B.h>9) K.B.h+=6; \
    R->A=(K.B.l&0x0F)|(K.B.h<<4); \
    R->P=(R->P&~C_FLAG)|(K.B.h>15? C_FLAG:0); \
  } \
  else \
  { \
    K.W=R->A+Rg+(R->P&C_FLAG); \
    R->P&=~(N_FLAG|V_FLAG|Z_FLAG|C_FLAG); \
    R->P|=(~(R->A^Rg)&(R->A^K.B.l)&0x80? V_FLAG:0)| \
          (K.B.h? C_FLAG:0)|ZNTable[K.B.l]; \
    R->A=K.B.l; \
  }

/* Warning! C_FLAG is inverted before SBC and after it */
#define M_SBC(Rg) \
  if(R->P&D_FLAG) \
  { \
    K.B.l=(R->A&0x0F)-(Rg&0x0F)-(~R->P&C_FLAG); \
    if(K.B.l&0x10) K.B.l-=6; \
    K.B.h=(R->A>>4)-(Rg>>4)-(K.B.l&0x10); \
    if(K.B.h&0x10) K.B.h-=6; \
    R->A=(K.B.l&0x0F)|(K.B.h<<4); \
    R->P=(R->P&~C_FLAG)|(K.B.h>15? 0:C_FLAG); \
  } \
  else \
  { \
    K.W=R->A-Rg-(~R->P&C_FLAG); \
    R->P&=~(N_FLAG|V_FLAG|Z_FLAG|C_FLAG); \
    R->P|=((R->A^Rg)&(R->A^K.B.l)&0x80? V_FLAG:0)| \
          (K.B.h? 0:C_FLAG)|ZNTable[K.B.l]; \
    R->A=K.B.l; \
  }

#define M_CMP(Rg1,Rg2) \
  K.W=Rg1-Rg2; \
  R->P&=~(N_FLAG|Z_FLAG|C_FLAG); \
  R->P|=ZNTable[K.B.l]|(K.B.h? 0:C_FLAG)
#define M_BIT(Rg) \
  R->P&=~(N_FLAG|V_FLAG|Z_FLAG); \
  R->P|=(Rg&(N_FLAG|V_FLAG))|(Rg&R->A? 0:Z_FLAG)

#define M_AND(Rg)	R->A&=Rg;M_FL(R->A)
#define M_ORA(Rg)	R->A|=Rg;M_FL(R->A)
#define M_EOR(Rg)	R->A^=Rg;M_FL(R->A)
#define M_INC(Rg)	Rg++;M_FL(Rg)
#define M_DEC(Rg)	Rg--;M_FL(Rg)

#define M_ASL(Rg)	R->P&=~C_FLAG;R->P|=Rg>>7;Rg<<=1;M_FL(Rg)
#define M_LSR(Rg)	R->P&=~C_FLAG;R->P|=Rg&C_FLAG;Rg>>=1;M_FL(Rg)
#define M_ROL(Rg)	K.B.l=(Rg<<1)|(R->P&C_FLAG); \
			R->P&=~C_FLAG;R->P|=Rg>>7;Rg=K.B.l; \
			M_FL(Rg)
#define M_ROR(Rg)	K.B.l=(Rg>>1)|(R->P<<7); \
			R->P&=~C_FLAG;R->P|=Rg&C_FLAG;Rg=K.B.l; \
			M_FL(Rg)

/** Reset6502() **********************************************/
/** This function can be used to reset the registers before **/
/** starting execution with Run6502(). It sets registers to **/
/** their initial values.                                   **/
/*************************************************************/
void Reset6502(M6502 * R)
{
	R->A = R->X = R->Y = 0x00;
	R->P = Z_FLAG | R_FLAG;
	R->S = 0xFF;
	R->PC.B.l = nes_read_byte(0xFFFC);
	R->PC.B.h = nes_read_byte(0xFFFD);
	R->ICount = R->IPeriod;
	R->IRequest = INT_NONE;
	R->AfterCLI = 0;
	R->TrapBadOps = 0;
}

/** Exec6502() ***********************************************/
/** This function will execute a single 6502 opcode. It     **/
/** will then return next PC, and current register values   **/
/** in R.                                                   **/
/*************************************************************/
inline word Exec6502(M6502 * R)
{
	register pair J, K;
	register byte I;

	I = Op6502(R->PC.W++);
	R->ICount -= Cycles[I];
	switch (I) {
#include "Codes.h"
	}

	/* We are done */
	return (R->PC.W);
}

/** Int6502() ************************************************/
/** This function will generate interrupt of a given type.  **/
/** INT_NMI will cause a non-maskable interrupt. INT_IRQ    **/
/** will cause a normal interrupt, unless I_FLAG set in R.  **/
/*************************************************************/
void Int6502(M6502 * R, byte Type)
{
	register pair J;

	if ((Type == INT_NMI) || ((Type == INT_IRQ) && !(R->P & I_FLAG))) {
		R->ICount -= 7;
		M_PUSH(R->PC.B.h);
		M_PUSH(R->PC.B.l);
		M_PUSH(R->P & ~B_FLAG);
		R->P &= ~D_FLAG;
		if (Type == INT_NMI)
			J.W = 0xFFFA;
		else {
			R->P |= I_FLAG;
			J.W = 0xFFFE;
		}
		R->PC.B.l = nes_read_byte(J.W++);
		R->PC.B.h = nes_read_byte(J.W);
	}
}

/** Run6502() ************************************************/
/** This function will run 6502 code until Loop6502() call  **/
/** returns INT_QUIT. It will return the PC at which        **/
/** emulation stopped, and current register values in R.    **/
/*************************************************************/
word Run6502(M6502 * R)
{
	register pair J, K;
	register byte I;
	register int r;

	for (;;) {

		I = Op6502(R->PC.W++);
		R->ICount -= Cycles[I];
		switch (I) {
#include "Codes.h"
		}
		if (R->ICount <= 0) {
			if (R->AfterCLI) {
				I = R->IRequest;
				R->ICount += R->IBackup - 1;
				R->AfterCLI = 0;
			} else {
				/*I=Loop6502(R);            /* Call the periodic handler */
				R->ICount = R->IPeriod;	/* Reset the cycle counter   */
			}
			if (I == INT_QUIT)
				return (R->PC.W);	/* Exit if INT_QUIT     */
			if (I)
				Int6502(R, I);	/* Interrupt if needed  */
		}
	}

	/* Execution stopped */
	return (R->PC.W);
}
