.data
.align 2
	.comm __pc, 2, 2
	.comm __a, 1, 1
	.comm __x, 1, 1
	.comm __y, 1, 1
	.comm __sp, 1, 1
	.comm __p, 1, 1
	.comm __cycles, 4, 4
.text 
.include "asmcpu/macros.h"
.include "asmcpu/instr.h"
.align 2
.include "asmcpu/asmtables.h"

!!!!!! CPU CORE
	.globl _asm_cpu_reset
	.type _asm_cpu_reset,@function
_asm_cpu_reset:
	mov.l	r14,@-r15
	sts.l	pr,@-r15

	mov	r15, r14
	! get PC reset address
	mov.l	.acr_pcif,r1
	mov.l	.acr_reset_vector, r4
	jsr	@r1
	nop

	mov.l	.acr_pc, r2
	mov.w	r0, @r2

	mov	#255, r1
	mov.l	.acr_sp, r2
	mov.b	r1, @r2

	mov	#64, r1
	mov.l	.acr_p, r2
	mov.b	r1, @r2

	mov	#0,r1
	mov.l	.acr_x, r2
	mov.b	r1, @r2

	mov.l	.acr_y, r2
	mov.b	r1, @r2

	mov.l	.acr_a, r2
	mov.b	r1, @r2

	mov	r14, r15
	lds.l	@r15+,pr	
	rts
	mov.l	@r15+,r14
.tmp: .align 2
.acr_pc: .long __pc
.acr_sp: .long __sp
.acr_p:  .long __p
.acr_x:  .long __x
.acr_y:  .long __y
.acr_a:  .long __a
.acr_pcif: .long _nes_read_word
.acr_reset_vector: .long 65532

	.globl _asm_cpu_execute
	.type _asm_cpu_execute,@function
_asm_cpu_execute:
	mov.l	r14,@-r15	! save stack ptr
	mov.l	r13,@-r15	! r13 will be used for cycle count
	mov.l	r12,@-r15	! r12  "   "   "    "  PC
	mov.l	r11,@-r15	! r11         ""       A
	mov.l	r10,@-r15	! r10         ""       X
	mov.l	r9, @-r15	! r9          ""       Y
	mov.l   r8, @-r15       ! r8          ""       SR
	sts.l	pr, @-r15
	mov	r15, r14

	! code right here should reset r13(cyclecount) to 0
	mov	#0, r13
	extu.b  r13, r13

	! get the PC from memory
	mov.l	.vpc, r1
	mov.w	@r1, r12
	extu.w  r12, r12

	! get A from memory
	mov.l	.va, r1
	mov.b	@r1, r11
	extu.b	r11, r11

	! get X
	mov.l	.vx, r1
	mov.b	@r1, r10
	extu.b	r10, r10

	! get Y
	mov.l	.vy, r1
	mov.b	@r1, r9
	extu.b	r9, r9

	! get P
	mov.l	.vp, r1
	mov.b	@r1, r8
	extu.b	r8, r8

_asm_cpu_execute_opcode:
!	mov.l	.j0, r1
!	jsr	@r1
!	mov	r12, r4

	mov.l	.vpc, r1
	mov.w	r12, @r1

	mov.l	.i2, r1		! nes_read_byte(pc)
	jsr	@r1		! |
	mov	r12, r4		! `---
	extu.b	r0, r0

	mov.l	.i1, r2
	mov.b	@(r0,r2), r1
	add	r1, r13		! this increases cycle count

	shll2	r0
	mov.l	.i0, r2
	mov.l	@(r0,r2), r1
	jmp	@r1		! execute opcode
	add	#1, r12		! PC++
_asm_cpu_execute_opcode_done:
!	mov.l	.pmf, r2
!	jsr	@r2
!	nop
_asm_cpu_add_cycles:
	! the add_cycles part gets done above, I should remove 
	! this label
_asm_cpu_cycles_done:
	mov	#113,r3

	cmp/ge  r3, r13
	bt	_asm_cpu_do_interrupt
	mov.l	.i3, r2
	jmp	@r2
	nop
_asm_cpu_do_interrupt:
	mov.l	.i6, r2	! call asm_cpu_interrupt
	jsr	@r2
	mov	#0, r13	! reset cycle counter
	
	tst	#1, r0
	bt 0f		! if result of asm_cpu_interrupt is *1* then NMI occured

	mov	r12,r4
	mov.l	.stwb, r2 ! stack_push((PC & 0xFF00) >> 8);
	jsr	@r2
	shlr8	r4	

	mov	r12,r4
	mov	#255,r3
	extu.b	r3, r3
	mov.l	.stwb, r2
	jsr	@r2	  ! stack_push(PC & 0xFF);
	and	r3, r4
	
	mov.l	.stwb, r2
	jsr	@r2	  ! stack_push(sr);
	mov	r8, r4

	mov.l	.nrw, r2
	mov.l	.interrupt_vector, r4
	jsr	@r2		! nes_read_word(interrupt_vector);
	nop

	mov	r0, r12		! pc = result
0:
	mov.l	.i5, r1
	mov.b	@r1, r2
	cmp/pl	r2
	bf	_asm_cpu_execute_done
	bra	_asm_cpu_execute_opcode
	nop
.align 2
.i5: .long _emulation_running
.i6: .long _asm_cpu_interrupt
.vcy: .long __cycles
.j0: .long _iopcode
.stwb: .long _asm_stack_push
.nrw: .long _nes_read_word
.interrupt_vector: .long 65530
.pmf: .long __print_msg_func
.align 2
_asm_cpu_execute_done:
	mov.l	.vpc,r2	! stores all our registers to memory
	mov.w	r12,@r2
	mov.l	.vx,r2
	mov.b	r10,@r2
	mov.l	.vy,r2
	mov.b	r9,@r2
	mov.l	.va,r2
	mov.b	r11,@r2
	mov.l	.vp,r2
	mov.b	r8,@r2

	mov	r14, r15	! restore old SP
	lds.l	@r15+, pr	! get old registers from stack
	mov.l	@r15+, r8
	mov.l	@r15+, r9
	mov.l	@r15+, r10
	mov.l	@r15+, r11
	mov.l	@r15+, r12
	mov.l	@r15+, r13
	rts
	mov.l	@r15+, r14	! restore stack ptr
.align 2
.i0: .long _asm_cpu_instructions
.i1: .long _asm_cpu_inst_cycles
.i2: .long _nes_read_byte
.i3: .long _asm_cpu_execute_opcode
.i4: .short 341	! cycles per scanline
.align 2
.vpc: .long __pc
.va: .long __a
.vx: .long __x
.vy: .long __y
.vp: .long __p


!!!!! STACK SHIT !!!!!!
_asm_stack_push:
	mov.l	r14, @-r15
	sts.l	pr,@-r15
	mov.l	r8,@-r15
	mov	r15, r14

	mov	r4, r5
	mov.l	.stackp0, r3

	mov.b	@r3, r8		! r8 now holds SP
	extu.b	r8, r8
	
	mov	r8, r4
	mov	#1, r3
	shll8	r3
	extu.w	r4, r4
	mov.l	.stackp1,r2
	jsr	@r2
	or	r3, r4
		
	add	#-1, r8
	mov.l	.stackp0, r3
	mov.b	r8, @r3

	mov	r15, r14
	mov.l	@r15+, r8
	lds.l	@r15+, pr
	rts
	mov.l	@r15+, r14
.align 2
.stackp0: .long __sp
.stackp1: .long _nes_write_byte

_asm_stack_pop:
	mov.l	r14, @-r15
	sts.l	pr,@-r15
	mov.l	r8,@-r15
	mov	r15, r14

	mov	r4, r5
	mov.l	.stacko0, r3

	mov.b	@r3, r8		! r8 now holds SP
	extu.b	r8, r8
	add	#1, r8		! SP++

	mov	r8, r4
	mov	#1, r3
	shll8	r3
	extu.w	r4, r4
	mov.l	.stacko1,r2
	jsr	@r2
	or	r3, r4
		
	mov.l	.stacko0, r3
	mov.b	r8, @r3

	mov	r15, r14
	mov.l	@r15+, r8
	lds.l	@r15+, pr
	rts
	mov.l	@r15+, r14
.align 2
.stacko0: .long __sp
.stacko1: .long _nes_read_byte

_asm_stack_pop_word:
	mov.l	r14, @-r15
	sts.l	pr,@-r15
	mov.l	r8,@-r15
	mov	r15, r14

	mov.l	.stackw0, r2
	jsr	@r2
	nop
	extu.b	r0, r0
	mov	r0, r8		! put 8bit result in r8

	mov.l	.stackw0, r2
	jsr	@r2
	nop
	extu.b	r0, r0
	shll8	r0		! shift left 8 bits
	or	r8, r0		! or with previous result

	mov	r15, r14
	mov.l	@r15+, r8
	lds.l	@r15+, pr
	rts
	mov.l	@r15+, r14
.align 2
.stackw0: .long _asm_stack_pop

!!!!!! INSTRUCTIONS !!!!!!


_asm_cpu_inst_asl:
	mov	#0xFE, r3
	extu.b	r3, r3
	and	r3, r8

	mov	#0x80, r3
	extu.b	r3, r3
	tst	r3, r11
	bt	0f

	mov	#1, r3
	or	r3, r8
0:
	shll	r11

	mov.l	.asl0, r4
	jmp	@r4
	extu.b	r11, r11
.align 2
.asl0: .long _asm_cpu_execute_opcode_done

_asm_cpu_inst_clc:
	mov	#254, r3	! bit 0 is C flag (0xFE is 254d)
	extu.b	r3, r3
	mov.l	.clc0, r4
	jmp	@r4
	and	r3, r8
.align 2
.clc0: .long _asm_cpu_execute_opcode_done

_asm_cpu_inst_cld:
	mov	#247, r3	! bit 3 is D flag (0xF7 is 247d)
	extu.b	r3, r3
	mov.l	.cld0, r4
	jmp	@r4
	and	r3, r8
.align 2
.cld0: .long _asm_cpu_execute_opcode_done

_asm_cpu_inst_cli:
	mov	#251, r3	! bit 2 is I flag (0xFB is 251d)
	extu.b	r3, r3
	mov.l	.cli0, r4
	jmp	@r4
	and	r3, r8
.align 2
.cli0: .long _asm_cpu_execute_opcode_done

_asm_cpu_inst_clv:
	mov	#191, r3	! bit 2 is I flag (0xBF is 191d)
	extu.b	r3, r3
	mov.l	.clv0, r4
	jmp	@r4
	and	r3, r8
.align 2
.clv0: .long _asm_cpu_execute_opcode_done

_asm_cpu_inst_dex:
	mov	#1, r3
	sub	r3, r10
	extu.b	r10, r10
	negative_zero_8bit r10
	mov.l	.dex0,r2
	jmp	@r2
	nop
.align 2
.dex0: .long _asm_cpu_execute_opcode_done

_asm_cpu_inst_dey:
	mov	#1, r3
	sub	r3, r9
	extu.b	r9, r9
	negative_zero_8bit r9
	mov.l	.dey0,r2
	jmp	@r2
	nop
.align 2
.dey0: .long _asm_cpu_execute_opcode_done

_asm_cpu_inst_inx:
	add	#1, r10
	extu.b	r10, r10
	negative_zero_8bit r10
	mov.l	.inx0,r2
	jmp	@r2
	nop
.align 2
.inx0: .long _asm_cpu_execute_opcode_done

_asm_cpu_inst_iny:
	add	#1, r9
	extu.b	r9, r9
	negative_zero_8bit r9
	mov.l	.iny0,r2
	jmp	@r2
	nop
.align 2
.iny0: .long _asm_cpu_execute_opcode_done

_asm_cpu_inst_lsr:
	mov	#0xfe, r3
	extu.b	r3, r3
	and	r3, r8

	mov	#1, r4
	tst	r4, r11
	bt	0f
	
	or	r4, r8	! set carry
	
0:
	mov.l	.lsr0, r2
	jmp	@r2
	shlr	r11

.align 2
.lsr0: .long _asm_cpu_execute_opcode_done

_asm_cpu_inst_nop:
	mov.l	.z0,r2
	jmp	@r2
	nop
.align 2
.z0: .long _asm_cpu_execute_opcode_done

_asm_cpu_inst_pha:
	mov.l	.pha1, r2
	jsr	@r2
	mov	r11, r4	! r11 is A
	
	mov.l	.pha0, r2
	jmp	@r2
	nop
.align 2
.pha0: .long _asm_cpu_execute_opcode_done
.pha1: .long _asm_stack_push

_asm_cpu_inst_php:
	mov.l	.php1, r2
	jsr	@r2
	mov	r8, r4	! r8 is P status
	
	mov.l	.php0, r2
	jmp	@r2
	nop
.align 2
.php0: .long _asm_cpu_execute_opcode_done
.php1: .long _asm_stack_push

_asm_cpu_inst_pla:
	mov.l	.pla1, r2
	jsr	@r2		! r0 = stack_pop()
	nop
	
	mov.l	.pla0, r2
	jmp	@r2
	mov	r0, r11		! r0 is return value / r11 is A
.align 2
.pla0: .long _asm_cpu_execute_opcode_done
.pla1: .long _asm_stack_pop

_asm_cpu_inst_plp:
	mov.l	.plp1, r2
	jsr	@r2		! r0 = stack_pop()
	nop
	
	mov.l	.plp0, r2
	jmp	@r2
	mov	r0, r8		! r0 is return value / r8 is SR
.align 2
.plp0: .long _asm_cpu_execute_opcode_done
.plp1: .long _asm_stack_pop

_asm_cpu_inst_rol:
	mov	r8, r2

	mov	#0xFE, r3
	extu.b	r3, r3
	and	r3, r8

	mov	r11, r1
	mov	#0x80, r4
	extu.b	r4, r4
	tst	r4, r11
	bt	0f

	mov	#1, r4
	or	r4, r8	! set carry
0:
	shll	r11
	extu.b	r11, r11

	mov	#1, r4
	tst	r4, r2	! was carry previously set?
	bt	1f	! no?
	or	r4, r11  ! yes, set bit 0 to 1
1:
	mov.l	.rol0, r2
	jmp	@r2
	nop
.align 2
.rol0: .long _asm_cpu_execute_opcode_done

_asm_cpu_inst_ror:
	mov	r8, r2

	mov	#0xFE, r3
	extu.b	r3, r3
	and	r3, r8

	mov	r11, r1
	mov	#0x01, r4
	tst	r4, r11
	bt	0f

	mov	#1, r4
	or	r4, r8	! set carry
0:
	shlr	r11
	
	mov	#1, r4
	tst	r4, r2	! was carry previously set?
	bt	1f	! no?
	mov	#0x80, r4
	extu.b	r4, r4
	or	r4, r11  ! yes, set bit 7 to 1
1:
	mov.l	.ror0, r2
	jmp	@r2
	nop
.align 2
.ror0: .long _asm_cpu_execute_opcode_done

_asm_cpu_inst_rti:
	mov.l	.rti1, r2	! pop()
	jsr	@r2
	nop
	mov	r0, r8		! P fromS

	mov.l	.rti2, r2	! pop_word()
	jsr	@r2
	nop
	mov	r0, r12		! PC fromS

	mov.l	.rti0, r2
	jmp	@r2
	nop
.align 2
.rti0: .long _asm_cpu_execute_opcode_done
.rti1: .long _asm_stack_pop
.rti2: .long _asm_stack_pop_word

_asm_cpu_inst_rts:
	mov.l	.rts1, r2	! pop_word()
	jsr	@r2
	nop
	add	#1, r0
	extu.w	r0, r0
	mov	r0, r12		! (PC fromS) + 1 -> PC

	mov.l	.rts0, r2
	jmp	@r2
	nop
.align 2
.rts0: .long _asm_cpu_execute_opcode_done
.rts1: .long _asm_stack_pop_word

_asm_cpu_inst_sec:
	mov	#1, r1	! bit 0 is carry
	mov.l	.sec0, r2
	jmp	@r2
	or	r1, r8
.align 2
.sec0: .long _asm_cpu_execute_opcode_done

_asm_cpu_inst_sed:
	mov	#0x08, r1	! bit 3 is decimal
	mov.l	.sed0, r2
	jmp	@r2
	or	r1, r8
.align 2
.sed0: .long _asm_cpu_execute_opcode_done

_asm_cpu_inst_sei:
	mov	#0x04, r1
	mov.l	.sei0, r2
	jmp	@r2
	or	r1, r8
.align 2
.sei0: .long _asm_cpu_execute_opcode_done

_asm_cpu_inst_tax:
	mov	r11, r10
	negative_zero_8bit r10

	mov.l	.tax0, r2
	jmp	@r2
	nop
.align 2
.tax0: .long _asm_cpu_execute_opcode_done

_asm_cpu_inst_tay:
	mov	r11, r9		! Y = A
	negative_zero_8bit r9	! set_flags(r9)

	mov.l	.tay0, r2
	jmp	@r2
	nop
.align 2
.tay0: .long _asm_cpu_execute_opcode_done

_asm_cpu_inst_tsx:
	mov.l	.tsx1, r2
	mov.b	@r2, r3
		
	mov	r3, r10		! X = SP
	negative_zero_8bit r10	! set_flags(r10)

	mov.l	.tsx0, r2
	jmp	@r2
	nop
.align 2
.tsx0: .long _asm_cpu_execute_opcode_done
.tsx1: .long __sp

_asm_cpu_inst_txa:
	mov	r10, r11	! A = X
	negative_zero_8bit r11	! set_flags(r11)

	mov.l	.txa0, r2
	jmp	@r2
	nop
.align 2
.txa0: .long _asm_cpu_execute_opcode_done

_asm_cpu_inst_txs:
	mov.l	.txs1, r2
	mov.b	r10, @r2	! SP = X

	mov.l	.txs0, r2
	jmp	@r2
	nop
.align 2
.txs0: .long _asm_cpu_execute_opcode_done
.txs1: .long __sp

_asm_cpu_inst_tya:
	mov	r9, r11	! A = Y
	negative_zero_8bit r11	! set_flags(r11)

	mov.l	.tya0, r2
	jmp	@r2
	nop
.align 2
.tya0: .long _asm_cpu_execute_opcode_done

.include "asmcpu/immediate.s"
.include "asmcpu/zeropage.s"
.include "asmcpu/zeropagex.s"
.include "asmcpu/absolute.s"
.include "asmcpu/absolutex.s"
.include "asmcpu/absolutey.s"
.include "asmcpu/piindirecty.s"
.include "asmcpu/pcrel.s"

!!!!!!! UNKNOWN !!!!!!!!
_asm_cpu_inst_unknown:
	mov.l	.upc,r2
	mov.w	r12,@r2
	mov.l	.ux,r2
	mov.b	r10,@r2
	mov.l	.uy,r2
	mov.b	r9,@r2
	mov.l	.ua,r2
	mov.b	r11,@r2
	mov.l	.up,r2
	mov.b	r8,@r2

	mov.l	.u2, r1
	shlr2	r0
	mov	r0, r4
	jsr	@r1
	mov	r12, r5


	mov.l	.u0,r2
	jmp	@r2
	nop
.align 2
!.u0: .long _asm_cpu_execute_opcode_done
.u0: .long _loop_forever
.u2: .long _errmsg
.ux: .long __x
.uy: .long __y
.ua: .long __a
.up: .long __p
.upc: .long __pc

.align 2
_loop_forever:
	mov.l	.lf0, r1
	jsr	@r1
	nop

	mov	#0,r1
	mov.l	.lf3,r3
	mov.b	@r3,r2
	cmp/pl	r2
	bf	_lf_getout

	mov.l	.lf1,r1
	jmp	@r1
	nop
_lf_getout:
	mov.l	.lf2,r1
	jmp	@r1
	nop

.align 2
.lf0: .long _gleam_poll_events
.lf1: .long _loop_forever
.lf2: .long _asm_cpu_execute_done
.lf3: .long _emulation_running



__print_msg_func:
	mov.l	r14, @-r15
	sts.l	pr, @-r15
	mov	r15, r14

	mov.l	.pmf_pc, r2
	mov.w	r12, @r2

	mov.l	.pmf_a, r2
	mov.b	r11, @r2

	mov.l	.pmf_x, r2
	mov.b	r10, @r2

	mov.l	.pmf_y, r2
	mov.b	r9, @r2

	mov.l	.pmf_p, r2
	mov.b	r8, @r2

	mov.l	.pmf_i, r2
	jsr	@r2
	nop

	mov	r14, r15
	lds.l	@r15+, pr
	rts
	mov.l	@r15+, r14
.pmf_pc: .long __pc
.pmf_a: .long __a
.pmf_x: .long __x
.pmf_y: .long __y
.pmf_p: .long __p
.pmf_i: .long _printstatus
