#ifdef TARGET_DC
#include <stdarg.h>

#include "emulate.h"
#include "fileio.h"
#include "dream.h"
#include "libc.h"
#include "dc.h"
#include "font.h"

#include "dc_sound/dcs_regs.h"
#include "dc_sound/dcsmain.h"

#include "vmsfs.h"

static nes_file gbl_nes_file;
static char gbl_char_data[1024 * 512];

void *fptr = (void *) NULL;

u16 dc_dbl_buf[640 * 240];
u16 option_buf[640 * 240];
u16 dc_colors[256];
u32 option_fullscreen = 0;
u32 option_frameskip = 0;
u32 in_option_screen = 0;
u32 use_alpha_blending = 0;
u16 *alpha_blending_ptr = (u32 *) NULL;

extern u16 vblcount;

void dc_gleam_sound_init();

void fflush(FILE * s)
{
}

FILE *fopen(char *fname, char *mode)
{
	int i;

	if (mode[0] == 'r' && mode[1] == 'b') {
		i = dc_gd_open(fname, O_RDONLY);
		return (FILE *) (i + 1);
	} else
		fprintf(stdout, "wasn't readonly\n");
	return 0;
}

void puts(char *str)
{
#ifndef NO_OUTPUT
	dc_serial_printf(str);
#endif
	return;
}

int fprintf(FILE * file, char *format, ...)
{
#ifndef NO_OUTPUT
	char buffer[2048];
	va_list args;

	va_start(args, format);
	vsprintf(buffer, format, args);
	va_end(args);

	dc_serial_printf("%s", buffer);
#endif
	return 0;
}

int printf(char *format, ...)
{
	return 0;
}

void signal(int signo, void (*func) ())
{
	return;
}

int fwrite(char *buf, int count, int nmemb, FILE * out)
{
	return nmemb;
}

int fread(char *buf, int count, int nmemb, FILE * in)
{
	int i, r;
	r = dc_gd_read(((int) in) - 1, buf, count * nmemb);
	if (r != (count * nmemb))
		return (r / count);
	if (r == 0)		/* eof */
		;
	return (r / count);
}

int fclose(FILE * file)
{
	dc_gd_close(((int) file) - 1);
	return 0;
}

void exit(int exitcode)
{
	/* system bios call - return to menu */
	(*(void (**)()) 0x8c0000e0) (1);
}

int fseek(FILE * file, int offset, int pos)
{
	return dc_gd_lseek(((int) file) - 1, offset, pos);
}

int ftell(FILE * file)
{
	return dc_gd_tell(((int) file) - 1);
}

void rewind(FILE * file)
{
	fseek(file, 0, SEEK_SET);
	return;
}

void *malloc(int size)
{
	static int call = 0;
	call++;
	if (size == 0) {
		call = 0;
		return 0;
	}
	if (call == 1)
		return (void *) &gbl_nes_file;
	else if (call == 2)
		return (void *) gbl_char_data;
}

void vram_blit()
{				/* not needed */
	video_memcpy64(vram_s, dc_dbl_buf, 4800);
	return;
}

int time(int whence)
{
	int vals[2];

	vals[0] = *((volatile u32 *) (0xA0710000));
	vals[1] = *((volatile u32 *) (0xA0710004));

	return (((vals[0] & 0x0000FFFF) << 16) | (vals[1] & 0x0000FFFF));
}

/* The following function comes from an email
 * conversation with Marcus Comstedt.
 */
#define QACR0 (*(volatile unsigned int *)(void *)0xff000038)
#define QACR1 (*(volatile unsigned int *)(void *)0xff00003c)

inline void video_memcpy64(void *dest, void *src, int cnt)
{
	unsigned int *s = (unsigned int *) src;
	unsigned int *d = (unsigned int *) (void *)
	    (0xe0000000 | (((unsigned long) dest) & 0x03ffffc0));
	QACR0 = ((0xa5000000 >> 26) << 2) & 0x1c;
	QACR1 = ((0xa5000000 >> 26) << 2) & 0x1c;
	while (cnt--) {
		d[0] = *s++;
		d[1] = *s++;
		d[2] = *s++;
		d[3] = *s++;
	      asm("pref @%0": :"r"(s + 16));
		d[4] = *s++;
		d[5] = *s++;
		d[6] = *s++;
		d[7] = *s++;
	      asm("pref @%0": :"r"(d));
		d += 8;
		d[0] = *s++;
		d[1] = *s++;
		d[2] = *s++;
		d[3] = *s++;
	      asm("pref @%0": :"r"(s + 16));
		d[4] = *s++;
		d[5] = *s++;
		d[6] = *s++;
		d[7] = *s++;
	      asm("pref @%0": :"r"(d));
		d += 8;
	}
}

#ifndef vram_plot
void vram_plot(int x, int y, int color)
{
	vram_s[32 + (y * 640) + x] = dc_colors[color];
}
#endif

void vram_plot_line(int y, u8 * color)
{
	int i;
	unsigned short *ptr = &dc_dbl_buf[(y * 640)];

	if (option_fullscreen == 0)
		ptr = &ptr[32];

	for (i = 0; i < 256; i++) {
		*ptr = dc_colors[*color];
		ptr++;
		if (option_fullscreen && ((i % 4) == 0))
			*ptr++ = dc_colors[*color];
		color++;
	}
}

unsigned short alphacombine(unsigned short clr0, unsigned char a0,
			    unsigned short clr1, unsigned char a1)
{
	int r0 = (clr0 & 0x7800) >> 10;
	int g0 = (clr0 & 0x07c0) >> 5;
	int b0 = clr0 & 0x001f;
	int r1 = (clr1 & 0x7800) >> 10;
	int g1 = (clr1 & 0x07c0) >> 5;
	int b1 = clr1 & 0x001f;
	int r, g, b;

	r = (r0 * a0 / 256) + (r1 * a1 / 256);
	g = (g0 * a0 / 256) + (g1 * a1 / 256);
	b = (b0 * a0 / 256) + (b1 * a1 / 256);

	return (r << 10) | (g << 5) | b;
}

void vram_plot_raw(int x, int y, int color, u16 * buffer)
{
#ifdef ALPHA
	if (use_alpha_blending)
		buffer[(y * 640) + x] =
		    alphacombine(dc_colors[color], 128,
				 alpha_blending_ptr[(y * 640) + x], 128);
	else
#endif
		buffer[(y * 640) + x] = dc_colors[color];
}

void vram_plot_dbl()
{
	u32 i;

	for (i = 0; i < (640 * 240); i++)
		vram_s[i] = dc_dbl_buf[i];
}

void dc_init()
{
#ifndef NO_OUTPUT
	dc_serial_init();
	dc_serial_printf("DreamCast gleam! init.\n");
#endif
	dc_gd_init();
	dc_font_set(gleam_font, gleam_font_height);
	memset(dc_dbl_buf, 0, 640 * 240 * sizeof(u16));

	dc_gleam_sound_init();

	use_alpha_blending = 1;
	alpha_blending_ptr = dc_dbl_buf;
}

void set_color(int color, int r, int g, int b)
{
	r >>= 1;
	b >>= 1;
	g >>= 1;
	dc_colors[color] =
	    (b & 0x1f) | ((g & 0x1f) << 5) | ((r & 0x1f) << 10);
}

void dc_setup_joy(u8 port, u16 data, u8 ltrig, u8 rtrig)
{
	u8 *joy;

	if (!port)
		joy = &JOYPAD1;
	else
		joy = &JOYPAD2;

	if (!(data & 0x400))
		*joy |= 0x02;
	else
		*joy &= ~0x02;
	if (!(data & 0x04))
		*joy |= 0x01;
	else
		*joy &= ~0x01;
	if (!(data & 0x10))
		*joy |= 0x10;
	else
		*joy &= ~0x10;
	if (!(data & 0x20))
		*joy |= 0x20;
	else
		*joy &= ~0x20;
	if (!(data & 0x40))
		*joy |= 0x40;
	else
		*joy &= ~0x40;
	if (!(data & 0x80))
		*joy |= 0x80;
	else
		*joy &= ~0x80;
	if (!(data & 0x08))
		*joy |= 0x08;
	else
		*joy &= ~0x08;
	if (ltrig == 255)
		*joy |= 0x04;
	else
		*joy &= ~0x04;
	if (rtrig == 255 && in_option_screen == 0) {
		DCS_STATUS = DCS_CMD_STOP;

		switch (dc_gleam_option_screen(current_filename)) {
		case 0:
			break;
		case 1:
			emulation_running = 0;
			break;
		}
	}
}

int gleam_poll_events()
{
	u8 port, device = 0;

	for (port = 0; port < 2; port++) {
		char *res;
		u32 params[1];
		u32 func;

		write_belong(&params[0], MAPLE_FUNC_CONTROLLER);

		res =
		    dc_maple_docmd(port, device, MAPLE_COMMAND_GETCOND, 1,
				   params);

		if (*res == MAPLE_RESPONSE_DATATRF && res[3] >= 2 &&
		    read_belong((u32 *) (res + 4)) ==
		    MAPLE_FUNC_CONTROLLER)
			dc_setup_joy(port, *((u16 *) (res + 8)),	/* buttons */
				     *((u8 *) (res + 11)),	/* left trigger */
				     *((u8 *) (res + 10)));	/* right trigger */
	}

	return 0;
}

/* GUI stuff, palettes 73 and up are free */
void dc_hline(int x1, int y1, int x2, int c, u16 * buffer)
{
	int temp;
	if (x1 > x2) {
		temp = x1;
		x1 = x2;
		x2 = temp;
	}

	for (temp = x1; temp < x2; temp++) {
		vram_plot_raw(temp, y1, c, buffer);
	}
}

void dc_vline(int x1, int y1, int y2, int c, u16 * buffer)
{
	int temp;
	if (y1 > y2) {
		temp = y1;
		y1 = y2;
		y2 = temp;
	}

	for (temp = y1; temp < y2; temp++) {
		vram_plot_raw(x1, temp, c, buffer);
	}
}

void dc_box_fill(int x1, int y1, int x2, int y2, int c, u16 * buffer)
{
	int temp;

	if (x1 > x2) {
		temp = x1;
		x1 = x2;
		x2 = temp;
	}
	if (y1 > y2) {
		temp = y1;
		y1 = y2;
		y2 = temp;
	}

	for (temp = y1; temp < y2; temp++) {
		dc_hline(x1, temp, x2, c, buffer);
	}
}

void dc_draw_right_arrow(int x1, int y1, int c, u16 * buffer)
{
	static u8 arrow[] = {
		0, 0, 0, 0, 0, 1, 1, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 1, 1, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 1, 1, 0,
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		0, 0, 0, 0, 0, 0, 0, 1, 1, 0,
		0, 0, 0, 0, 0, 0, 1, 1, 0, 0,
		0, 0, 0, 0, 0, 1, 1, 0, 0, 0,
	};			/* 10 x 8 */
	int x, y;

	for (y = 0; y < 8; y++) {
		for (x = 0; x < 10; x++) {
			if (arrow[(y * 10) + x])
				vram_plot_raw(x1 + x, y1 + y, c, buffer);
		}
	}
}

void dc_gui_load_rom(char *buf)
{
	int toptobottom = 0;	/* variable that is used to skip top X files */
	int fcount = 0;
	int cpos = 0;		/* cursor position, 0 to fcount */
	int alldone = 0;
	int temp;
	int filecount = 0;

	DIR *dir;
	struct dirent *dirent;

	filecount = dc_count_files();

	for (; !alldone;) {
		int x_off, y_off;

		dc_box_fill(32, 24, 320 - 32, 220, 75, option_buf);
		dc_draw_string(32, 24, RGBS(255, 255, 255), "Load Rom...",
			       option_buf);

		dc_box_fill(40, 48, 40 + (8 * 25), 220 - 8, 73,
			    option_buf);

		fcount = (((220 - 8) - 48) - 1) / 16;

		x_off = 40 + 8 + 10;
		y_off = 49;

		/* calculate what 9 files need to be listed */
		dir = dc_gd_opendir("/");
		if (!dir) {
			dc_error_screen();
			cpos = 0;
			toptobottom = 0;
			filecount = dc_count_files();
			dc_vid_clear(0, 0, 0);
			continue;
		}

		for (temp = 0; temp < toptobottom; temp++)
			dirent = dc_gd_readdir(dir);
		for (temp = 0; temp < fcount; temp++) {
			if ((toptobottom + temp) == filecount)
				break;

			dc_box_fill(40, y_off, 40 + (8 * 25), y_off + 16,
				    73, option_buf);

			if (cpos == temp) {
				dc_draw_right_arrow(x_off - 16, y_off + 4,
						    77, option_buf);
			}
			dirent = dc_gd_readdir(dir);
			if (dirent->d_size) {
				dc_draw_string(x_off, y_off,
					       RGBS(255, 255, 255),
					       dirent->d_name, option_buf);
				y_off += 16;
			} else
				temp--;
		}
		dc_gd_closedir(dir);

		video_memcpy64(vram_s, option_buf, 4800);

		{
			u8 oldjoy;
			u8 done = 0;

			while (!done) {
				dir = dc_gd_opendir("/");
				if (!dir) {
					done = 1;
					continue;
				}
				dc_gd_closedir(dir);

				gleam_poll_events();	/* update joypad */

				if (JOYPAD1 & 0x20) {
					if (JOYPAD1 & 0x02) {
						if ((cpos + toptobottom +
						     fcount) < filecount)
							toptobottom +=
							    fcount;
					} else {
						if ((cpos + 1) == fcount) {
							if (((toptobottom +
							      fcount) +
							     1) <=
							    filecount)
								toptobottom++;	/* this will cause errors */
						} else {
							if ((cpos +
							     toptobottom +
							     1) <
							    filecount)
								cpos++;
						}
					}
					done = 1;
				} else if (JOYPAD1 & 0x10) {
					if (JOYPAD1 & 0x02) {
						if (toptobottom >= fcount)
							toptobottom -=
							    fcount;
					} else {
						if (cpos == 0) {
							if (toptobottom >
							    0)
								toptobottom--;
						} else
							cpos--;
					}
					done = 1;
				} else if (JOYPAD1 & 0x01) {	/* done */
					done = 1;
					alldone = 1;
				} else if (JOYPAD1 & 0x08) {
					done = 1;
					alldone = 1;
				}
			}
		}

	}
	dc_draw_string((320 - 32) - 80, 24, RGBS(255, 255, 255),
		       "Loading...", vram_s);
	dir = dc_gd_opendir("/");
	for (temp = 0; temp < toptobottom; temp++)
		dirent = dc_gd_readdir(dir);
	for (temp = 0; temp < fcount; temp++) {
		dirent = dc_gd_readdir(dir);

		if (cpos == temp) {
			if (dirent->d_size)
				strcpy(buf, dirent->d_name);
		}
		if (!dirent->d_size)
			temp--;
	}
	dc_gd_closedir(dir);
	return;
}

void dc_intro_screen()
{
	int done = 0;

	dc_vid_clear(0, 0, 0);

	dc_draw_string(20, 20, RGBS(232, 232, 232),
		       "Despite the previous screen", vram_s);
	dc_draw_string(20, 36, RGBS(232, 232, 232),
		       "that is forced to display,", vram_s);
	dc_draw_string(20, 52, RGBS(255, 255, 255),
		       "gleam! is not affiliated with", vram_s);
	dc_draw_string(20, 68, RGBS(255, 255, 255),
		       "Sega or any of Sega's subsidiaries.", vram_s);

	dc_draw_string(20, 100, RGBS(232, 232, 232),
		       "Make sure a NES cdrom is inserted...", vram_s);
	dc_draw_string(20, 116, RGBS(232, 232, 232),
		       "Press Start to begin...", vram_s);

	while ((JOYPAD1 & 0x08))
		gleam_poll_events();

	while (!done) {
		gleam_poll_events();
		if (JOYPAD1 & 0x08)
			done = 1;
	}
}

void dc_error_screen()
{
	DIR *d;
	int done = 0;

	dc_vid_clear(0, 0, 0);
	dc_draw_string(20, 20, RGBS(255, 255, 255),
		       "Cannot read from CDROM drive...", vram_s);
	dc_draw_string(20, 36, RGBS(255, 255, 255),
		       "Please insert a CDROM containing", vram_s);
	dc_draw_string(20, 52, RGBS(255, 255, 255), "       NES roms...",
		       vram_s);

	while (!done) {
		d = dc_gd_opendir("/");
		if (d)
			done = 1;
	}
	dc_gd_closedir(d);
}

void dc_vmu_error_screen()
{

	dc_vid_clear(0, 0, 0);
	dc_draw_string(20, 20, RGBS(255, 255, 255),
		       "Cannot locate a Memory Card", vram_s);
	dc_draw_string(20, 28, RGBS(255, 255, 255),
		       "With at least 17 blocks free", vram_s);
	dc_draw_string(20, 44, RGBS(255, 255, 255),
		       "Please insert a Memory Card", vram_s);
	dc_draw_string(20, 60, RGBS(255, 255, 255),
		       " into any DC controller slot...", vram_s);

	dc_draw_string(20, 90, RGBS(255, 255, 255),
		       "Press START to return...", vram_s);

	while (!(JOYPAD1 & 0x08))
		gleam_poll_events();

	return;
}

int dc_count_files()
{
	int i = 0;
	DIR *d;
	struct dirent *de;

	d = dc_gd_opendir("/");
	if (!d)
		dc_error_screen();
	for (;;) {
		de = dc_gd_readdir(d);
		if (de) {
			if (de->d_size)
				i++;
		} else
			break;
	}
	dc_gd_closedir(d);
	return i;
}

void dc_gleam_sound_init()
{
	/* Pull reset low on ARM */
	*((volatile unsigned long *) (void *) 0xa0702c00) |= 1;

	memset4((void *) DCS_SOUNDBASE, 0, DCS_SOUNDSIZE);

	memcpy4((void *) DCS_SOUNDBASE, dcs_code, dcs_code_size);

	/* start execution */
	*((volatile unsigned long *) (void *) 0xa0702c00) &= ~1;

	while (DCS_STATUS != 1) {
		fprintf(stdout, "%p is %08X\n", &DCS_STATUS, DCS_STATUS);
	}
	DCS_STATUS = 0;

	fprintf(stdout, "%p is now %08X\n", &DCS_STATUS, DCS_STATUS);

}

/* OPTION SCREEN */
int dc_gleam_option_screen(char *argument)
{
	int done = 0;
	int pos = 0;
	int rom_loaded = 0;
	int reason_to_reset = 0;
	int saveram = 0;

	char *cpustrings[] =
	    { "CPU core...M6502", "CPU core...ASMCPU",
   "CPU core...G6502" };
	char ostr[128];

	in_option_screen = 1;

	memcpy(option_buf, dc_dbl_buf, 640 * 240 * sizeof(u16));


	while (!done) {
		int checking_controller = 1;

		alpha_blending_ptr = dc_dbl_buf;

		dc_box_fill(32, 24, 320 - 32, 220, 75, option_buf);
		dc_draw_string(32, 24, RGBS(255, 255, 255), "Options...",
			       option_buf);

		dc_box_fill(40, 48, 320 - 40, 220 - 8, 73, option_buf);

		if (rom_loaded == 0)
			dc_draw_string(58, 49, RGBS(255, 255, 255),
				       "Load Rom...", option_buf);
		else
			dc_draw_string(58, 49, RGBS(255, 255, 255),
				       "Load Rom...READY", option_buf);

		switch (saveram) {
		case 0:
			dc_draw_string(58, 65, RGBS(255, 255, 255),
				       "Save RAM...", option_buf);
			break;
		case 1:
			dc_draw_string(58, 65, RGBS(255, 255, 255),
				       "Save RAM...Saved", option_buf);
			break;
		}

		dc_draw_string(58, 81, RGBS(255, 255, 255),
			       cpustrings[(int) current_cpu_type],
			       option_buf);

		if (option_fullscreen)
			dc_draw_string(58, 97, RGBS(255, 255, 255),
				       "Full Screen...ON", option_buf);
		else
			dc_draw_string(58, 97, RGBS(255, 255, 255),
				       "Full Screen...OFF", option_buf);

		sprintf(ostr, "Frame Skip...%d\0", option_frameskip);
		dc_draw_string(58, 113, RGBS(255, 255, 255), ostr,
			       option_buf);

		if (emulation_running == 1 && reason_to_reset == 0)
			dc_draw_string(58, 129, RGBS(255, 255, 255),
				       "Return", option_buf);
		else
			dc_draw_string(58, 129, RGBS(255, 255, 255),
				       "Start", option_buf);

		dc_draw_right_arrow(42, (49 + (pos * 16)) + 4, 77,
				    option_buf);

		video_memcpy64(vram_s, option_buf, 4800);

		while (checking_controller) {
			u8 oldjoy = JOYPAD1;

			while (oldjoy == JOYPAD1)
				gleam_poll_events();

			if (JOYPAD1 & 0x01) {
				switch (pos) {
				case 0:
					dc_gui_load_rom(argument);
					rom_loaded = 1;
					reason_to_reset = 1;
					break;
				case 1:
					dc_draw_string(58, 65,
						       RGBS(255, 255, 255),
						       "Save RAM...Saving",
						       option_buf);
					saveram =
					    dc_gleam_saveram_option();
					break;
				case 2:
					current_cpu_type =
					    dc_gleam_cpu_core_option();
					reason_to_reset = 1;
					break;
				case 3:
					option_fullscreen =
					    dc_gleam_full_screen_option();
					if (!option_fullscreen) {	/* clear borders */
						dc_box_fill(0, 0, 32, 239,
							    73,
							    dc_dbl_buf);
						dc_box_fill(0 + 32 + 256,
							    0,
							    0 + 32 + 256 +
							    32, 239, 73,
							    dc_dbl_buf);
					}
					break;
				case 4:
					option_frameskip =
					    dc_gleam_frameskip_option();
					vblcount = 0;
					break;
				case 5:	/* return / start */
					in_option_screen = 0;
					if (emulation_running == 0)
						return 1;	/* for now */
					else {
						if (reason_to_reset) {
							vblcount = 0;
							return 1;

						} else
							return 0;
					}
					break;
				default:
					continue;
				}
				checking_controller = 0;
			} else if (JOYPAD1 & 0x02) {
				in_option_screen = 0;
				return 0;
			} else if (JOYPAD1 & 0x20) {	/* down */
				if (pos == 5)
					continue;
				pos++;
				checking_controller = 0;
			} else if (JOYPAD1 & 0x10) {	/* up */
				if (pos == 0)
					continue;
				pos--;
				checking_controller = 0;
			}
		}
	}
}

cpu_type dc_gleam_cpu_core_option()
{
	const int CPUCOUNT = 3;

	int pos = (int) current_cpu_type;
	int done = 0;

	if (JOYPAD1 & 0x01) {
		while (JOYPAD1 & 0x01)
			gleam_poll_events();
	}

	alpha_blending_ptr = option_buf;

	while (!done) {
		int cc = 1;
		int xpos, ypos;

		dc_box_fill(32 + 8 + 10 + (11 * 8), 49 + (2 * 16),
			    32 + 8 + 10 + (14 * 8) + 16 + (8 * 6) + 10,
			    49 + (2 * 16) + (16 * 4), 77, option_buf);
		dc_box_fill(32 + 8 + 10 + (11 * 8) + 8, 49 + (2 * 16) + 8,
			    32 + 8 + 10 + (14 * 8) + 8 + (8 * 6) + 10,
			    49 + (3 * 16) + (16 * 2) + 8, 73, option_buf);
		dc_draw_string(32 + 8 + 10 + (11 * 8) + 8 + 10,
			       49 + (2 * 16) + 8 + 0, RGBS(255, 255, 255),
			       "M6502", option_buf);
		dc_draw_string(32 + 8 + 10 + (11 * 8) + 8 + 10,
			       49 + (2 * 16) + 8 + 16, RGBS(255, 255, 255),
			       "ASMCPU", option_buf);
		dc_draw_string(32 + 8 + 10 + (11 * 8) + 8 + 10,
			       49 + (2 * 16) + 8 + 32, RGBS(255, 255, 255),
			       "G6502", option_buf);


		xpos = 32 + 8 + 10 + (11 * 8) + 8;
		ypos = 49 + (2 * 16) + 9 + 4;
		ypos += (16 * pos);

		dc_draw_right_arrow(xpos, ypos, 75, option_buf);

		video_memcpy64(vram_s, option_buf, 4800);

		while (cc == 1) {
			u8 oj = JOYPAD1;

			while (oj == JOYPAD1)
				gleam_poll_events();

			if (JOYPAD1 & 0x01) {
				return (cpu_type) pos;
			} else if (JOYPAD1 & 0x10) {
				if (pos > 0)
					pos--;
				cc = 0;
			} else if (JOYPAD1 & 0x20) {
				if (pos < CPUCOUNT)
					pos++;
				cc = 0;
			}
		}
	}
}

int dc_gleam_full_screen_option()
{
	int pos = option_fullscreen ^ 1;
	int done = 0;

	if (JOYPAD1 & 0x01) {
		while (JOYPAD1 & 0x01)
			gleam_poll_events();
	}

	alpha_blending_ptr = option_buf;

	while (!done) {
		int cc = 1;

		dc_box_fill(32 + 8 + 10 + (14 * 8), 49 + (3 * 16),
			    32 + 8 + 10 + (14 * 8) + 16 + (8 * 3) + 10,
			    49 + (3 * 16) + (16 * 3), 77, option_buf);
		dc_box_fill(32 + 8 + 10 + (14 * 8) + 8, 49 + (3 * 16) + 8,
			    32 + 8 + 10 + (14 * 8) + 8 + (8 * 3) + 10,
			    49 + (3 * 16) + (16 * 2) + 8, 73, option_buf);
		dc_draw_string(32 + 8 + 10 + (14 * 8) + 8 + 10,
			       49 + (3 * 16) + 8 + 0, RGBS(255, 255, 255),
			       "On", option_buf);
		dc_draw_string(32 + 8 + 10 + (14 * 8) + 8 + 10,
			       49 + (3 * 16) + 8 + 16, RGBS(255, 255, 255),
			       "Off", option_buf);

		if (pos == 0)
			dc_draw_right_arrow(32 + 8 + 10 + (14 * 8) + 8,
					    49 + (3 * 16) + 9 + 4, 75,
					    option_buf);
		else
			dc_draw_right_arrow(32 + 8 + 10 + (14 * 8) + 8,
					    49 + (3 * 16) + 9 + 20, 75,
					    option_buf);

		video_memcpy64(vram_s, option_buf, 4800);

		while (cc == 1) {
			u8 oj = JOYPAD1;

			while (oj == JOYPAD1)
				gleam_poll_events();

			if (JOYPAD1 & 0x01)
				return !pos;
			else if (JOYPAD1 & 0x10 && pos == 1) {
				pos = 0;
				cc = 0;
			} else if (JOYPAD1 & 0x20 && pos == 0) {
				pos = 1;
				cc = 0;
			}
		}
	}
}

int dc_gleam_frameskip_option()
{
	int pos = option_frameskip;
	char str[4];
	int done = 0;

	if (JOYPAD1 & 0x01) {
		while (JOYPAD1 & 0x01)
			gleam_poll_events();
	}

	alpha_blending_ptr = option_buf;


	while (!done) {
		int cc = 1;

		sprintf(str, "%2d\0", pos);

		dc_box_fill(32 + 8 + 10 + (13 * 8), 49 + (4 * 16),
			    32 + 8 + 10 + (13 * 8) + 16 + (8) + 10,
			    49 + (4 * 16) + (16 * 2), 77, option_buf);
		dc_box_fill(32 + 8 + 10 + (13 * 8) + 8, 49 + (4 * 16) + 8,
			    32 + 8 + 10 + (13 * 8) + 8 + (8) + 10,
			    49 + (4 * 16) + (16 * 1) + 8, 73, option_buf);
		dc_draw_string(32 + 8 + 10 + (13 * 8) + 9,
			       49 + (4 * 16) + 8 + 0, RGBS(255, 255, 255),
			       str, option_buf);

		video_memcpy64(vram_s, option_buf, 4800);

		while (cc == 1) {
			u8 oj = JOYPAD1;

			while (oj == JOYPAD1)
				gleam_poll_events();

			if (JOYPAD1 & 0x01)
				return pos;
			else if (JOYPAD1 & 0x40 && pos > 0) {
				if (JOYPAD1 & 0x02) {
					if (pos < 10)
						pos = 0;
					else
						pos -= 10;
				} else
					pos--;
				cc = 0;
			} else if (JOYPAD1 & 0x80 && pos < 99) {
				if (JOYPAD1 & 0x02) {
					if (pos > 89)
						pos = 99;
					else
						pos += 10;
				} else
					pos++;
				cc = 0;
			}
		}
	}
}

void timestamp_it(struct timestamp *tstamp, unsigned long timesig);

void make_filename(struct vms_file_header *g_fileheader, char *new,
		   char *fname)
{
	char filename[128];
	char *ptr;

	strcpy(filename, fname);
	filename[13] = '\0';
	ptr = (char *) strchr(filename, '.');
	if (!ptr)
		ptr = filename;
	*ptr++ = '.';
	*ptr++ = 'G';
	*ptr++ = 'L';
	*ptr++ = 'M';

	strcpy(new, filename);

	strcpy(g_fileheader->shortdesc, "NES SAVERAM");
	sprintf(g_fileheader->longdesc, "%s - gleam! Save RAM", fname);
	sprintf(g_fileheader->id, "%c%c%c%c              ", 0xde, 0xad,
		0xbe, 0xef);
}


int dc_gleam_saveram_option()
{
	int port;
	int unit;

	struct vmsinfo g_vmsinfo;
	struct superblock g_superblock;
	struct vms_file_header g_fileheader;
	struct timestamp g_timestamp;

	char filename[128];


	memset(&g_fileheader, 0, sizeof(struct vms_file_header));
	make_filename(&g_fileheader, filename, current_filename);
	g_fileheader.numicons = 0;


	timestamp_it(&g_timestamp, time(NULL));


	for (port = 0; port < 2; port++) {	/* gleam only uses controllers A & B */
		for (unit = 0; unit < 6; unit++) {
			int i = (port * 6) + unit;
			int j;

			j = vmsfs_check_unit(i, 0, &g_vmsinfo);

			if (j == 1) {
				/* Found, save and return */
				if (vmsfs_get_superblock
				    (&g_vmsinfo, &g_superblock)) {
					/* memory card has been formatted, lets open and write a file */
					vmsfs_beep(&g_vmsinfo, 1);

					if (vmsfs_create_file
					    (&g_superblock, filename,
					     &g_fileheader, NULL, NULL,
					     nes_sram, 0x2000,
					     &g_timestamp)) {
						vmsfs_beep(&g_vmsinfo, 0);
						return 1;
					} else {
						vmsfs_beep(&g_vmsinfo, 0);
					}

				}

			}
		}
	}

	dc_vmu_error_screen();

	return 0;
}

int dc_gleam_loadram_option()
{
	struct vmsinfo g_vmsinfo;
	struct superblock g_superblock;
	struct vms_file g_file;
	struct vms_file_header g_fileheader;
	int port, unit;
	char filename[128];


	make_filename(&g_fileheader, filename, current_filename);

	for (port = 0; port < 2; port++) {
		for (unit = 0; unit < 6; unit++) {
			int i = (port * 6) + unit;
			int j;

			j = vmsfs_check_unit(i, 0, &g_vmsinfo);

			if (j == 1) {
				/* Found, load nes_sram and return */
				if (vmsfs_get_superblock
				    (&g_vmsinfo, &g_superblock)) {
					if (vmsfs_open_file
					    (&g_superblock, filename,
					     &g_file)) {
						vmsfs_beep(&g_vmsinfo, 1);
						vmsfs_read_file(&g_file,
								nes_sram,
								0x2000);
						vmsfs_beep(&g_vmsinfo, 0);

						return 1;
					}
				}
			}
		}
	}
	return 0;

}

unsigned char daycounts[12] =
    { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

void timestamp_it(struct timestamp *tstamp, unsigned long timesig)
{
	short int year, month, day, wkday, hour, minute, second;

	for (year = 1950; timesig >= (24 * 60 * 60 * 365); year++) {
		if ((year % 4) == 0)
			timesig -= (24 * 60 * 60 * 366);
		else
			timesig -= (24 * 60 * 60 * 365);
	}

	for (month = 0; month < 11 && (timesig > (28 * (24 * 60 * 60)));
	     month++) {
		if ((month == 1) && ((year % 4) == 0)) {
			if (timesig < (29 * 24 * 60 * 60))
				break;
			timesig -= 29 * (24 * 60 * 60);
		} else {
			if (timesig < (daycounts[month] * 24 * 60 * 60))
				break;
			timesig -= daycounts[month] * (24 * 60 * 60);
		}
	}

	for (day = 0; day < 31 && timesig > (24 * 60 * 60); day++) {
		timesig -= (24 * 60 * 60);
	}

	for (hour = 0; hour < 24 && (timesig > (60 * 60)); hour++) {
		timesig -= (60 * 60);
	}

	for (minute = 0; minute < 60 && (timesig > 60); minute++) {
		timesig -= 60;
	}

	second = timesig;

	tstamp->year = year;
	tstamp->month = month + 1;
	tstamp->day = day + 1;
	tstamp->hour = hour;
	tstamp->minute = minute;
	tstamp->second = second;
	tstamp->wkday = day & 7;	/* i know this is wrong */
}

#endif
