#ifndef TARGET_DC

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <readline/readline.h>
#include <readline/history.h>


#include "gleam!.h"
#include "emulate.h"
#include "g6502.h"
#include "mem.h"


typedef struct _debug_cmd {
	s8 cmd;
	void (*func) (g6502_cpu *, s8 *);
	s8 *helptext;
} debug_cmd;

void debug_step(g6502_cpu *, s8 *);
void debug_regs(g6502_cpu *, s8 *);
void debug_quit(g6502_cpu *, s8 *);
void debug_help(g6502_cpu *, s8 *);

debug_cmd debug_commands[] = {
	{(s8) 's', debug_step, (s8 *) "Execute next instruction"},
	{(s8) 'r', debug_regs, (s8 *) "Show register information"},
	{(s8) 'q', debug_quit, (s8 *) "Quits the debugger"},
	{(s8) 'h', debug_help, (s8 *) "Gives help information"},
	{(s8) 0, debug_quit, (s8 *) NULL},
};


static int d_quit = 0;

void debug(g6502_cpu * cpu)
{
	s8 *in;
	char cmd, oldcmd = 0;
	u8 i;

	while (!d_quit) {
		in = (s8 *) readline("6502> ");
		cmd = *in;
		add_history((char *) in);

		if (cmd == 0 && oldcmd != 0) {
			cmd = oldcmd;
		} else {
			oldcmd = cmd;
		}

		for (i = 0; debug_commands[i].cmd; i++) {
			if (debug_commands[i].cmd == tolower(cmd))
				break;
		}
		if (debug_commands[i].cmd)
			debug_commands[i].func(cpu, in);
	}

}

void debug_step(g6502_cpu * a, s8 * b)
{
	g6502_step();
}

void debug_regs(g6502_cpu * a, s8 * b)
{
	fprintf(stdout,
		"PC: [%04X] SP: [%02X] SR: [%02X] X: [%02X] Y: [%02X] A: [%02X]\n",
		a->pc, a->sp, a->p, a->x, a->y, a->a);
}

void debug_quit(g6502_cpu * a, s8 * b)
{
	d_quit = 1;
}

void debug_help(g6502_cpu * a, s8 * b)
{
	u8 i;

	for (i = 0; debug_commands[i].cmd; i++) {
		fprintf(stdout, "%c - %s\n", debug_commands[i].cmd,
			debug_commands[i].helptext);
	}
}

#endif	/* TARGET_DC */
