#ifndef TARGET_DC
#  include <stdio.h>
#  include <stdlib.h>
#else
#  include "dream.h"
#  include "libc.h"
#  include "dc.h"
#  include "asmcpu.h"
#endif

#include "gleam!.h"
#include "fileio.h"
#include "emulate.h"
#include "mem.h"
#include "g6502.h"
#include "debug.h"
#include "tables.h"
#include "M6502.h"

#ifdef TARGET_DC
# include "dc_sound/dcs_regs.h"
#endif

u8 nes_ram[0x0800];
u16 nes_regs[8];
u8 nes_soundregs[0x16];
u8 nes_sram[0x2000];
u8 nes_vram[0x4000];
u8 nes_sprram[0x100];
u8 nes_joy[2];

u8 *prg_banks[4];

u8 *chr_rom_ptr = (u8 *) NULL;

u8 nes_trainer[512];

u8 nes_single_screen_mirroring = 0;
u8 emulation_running = 0;

extern u16 current_line;
extern unsigned long time_start;
extern int total_vblcount;

cpu_type current_cpu_type = CPU_M6502;

void nes_emulate(nes_file * nf)
{
	M6502 cpu;
	extern u8 irq_control, irq_count_down;

	fprintf(stdout, "-> nes_emulate()\n");

	memset(nes_ram, 0, 0x0800);
	memset(nes_regs, 0, 16);
	memset(nes_sram, 0, 0x2000);
	memset(nes_vram, 0, 0x4000);
	memset(nes_sprram, 0, 0x100);
	memset(nes_joy, 0, 2);

	current_line = 0;

	/* For DC vbl counter */
	time_start = time(NULL);
	total_vblcount = 0;
	/* .. */

	nes_mem_reset();

	if (nf->ucb[0] & 0x04) {
		memcpy(nes_trainer, nf->data, 512);
		memcpy(nf->data, &(nf->data[512]),
		       (nf->prg_count * 16384) + (nf->chr_count * 8192));
	}

	if (nf->chr_count)
		chr_rom_ptr = &(nf->data[nf->prg_count * 16384]);
	else
		chr_rom_ptr = (u8 *) NULL;

	nes_mapper_init(nf);

	emulation_running = 1;

#ifdef TARGET_DC
	dc_gleam_loadram_option();
#endif
	switch (current_cpu_type) {
	case CPU_M6502:
		cpu.IPeriod = 115;
		Reset6502(&cpu);

		for (; emulation_running;) {
			Exec6502(&cpu);
			if (cpu.ICount < 0) {
				if (next_line() == NMI_VECTOR) {
					if (PPUCNT1 & 0x80)
						Int6502(&cpu, INT_NMI);
					gleam_poll_events();
				} else {
/* 						if(irq_control) {
 						printf("here\n");
							if(--irq_count_down == 0)
								Int6502(&cpu, INT_IRQ);
						}*/
				}
				gleam_process_sound();
				cpu.ICount = cpu.IPeriod;
			}
		}

		break;
#ifdef TARGET_DC
	case CPU_ASM:
		asm_cpu_reset();
		/*dc_gleam_sound_init(); */
		asm_cpu_execute();
		break;
#endif
	case CPU_G6502:
		g6502_reset();
		while (g6502_step() && emulation_running) {
			if (gleam_cpu.cycle_count >= 341) {
				if (next_line() == NMI_VECTOR) {
					if (PPUCNT1 & 0x80)
						g_do_nmi();
					gleam_poll_events();
				}
				gleam_cpu.cycle_count = 0;
			}
		}
		break;
	default:
		break;
	}

	fprintf(stdout, "<- nes_emulate();\n");
}

void nes_set_rom_page32k(u8 prgrom)
{
	prg_banks[0] = &current_nes_file->data[prgrom * 0x8000];
	prg_banks[1] = prg_banks[0] + 0x2000;
	prg_banks[2] = prg_banks[1] + 0x2000;
	prg_banks[3] = prg_banks[2] + 0x2000;
}

void nes_set_rom_page(u8 bank, u8 prgrom)
{
	switch (bank) {
	case 0:
		prg_banks[0] = &current_nes_file->data[prgrom * 16384];
		prg_banks[1] = prg_banks[0] + 0x2000;
		break;
	case 1:
		prg_banks[2] = &current_nes_file->data[prgrom * 16384];
		prg_banks[3] = prg_banks[2] + 0x2000;
		break;
	}
}

void nes_set_rom_page8k(u8 bank, u8 prgrom)
{
	switch (bank) {
	case 0:
		prg_banks[0] = &current_nes_file->data[prgrom * 8192];
		break;
	case 1:
		prg_banks[1] = &current_nes_file->data[prgrom * 8192];
		break;
	case 2:
		prg_banks[2] = &current_nes_file->data[prgrom * 8192];
		break;
	case 3:
		prg_banks[3] = &current_nes_file->data[prgrom * 8192];
		break;
	}
}

void nes_set_chr_page8k(u8 chrrom)
{				/* 8k */
	if (chr_rom_ptr)
		memcpy(nes_vram, chr_rom_ptr + (chrrom * 0x2000), 0x2000);
}


void nes_set_chr_page(u8 bank, u8 chrrom)
{				/* banks are 4k in size */
	if (!chr_rom_ptr)
		return;

	switch (bank) {
	case 0:
		memcpy(nes_vram, chr_rom_ptr + (chrrom * 0x1000), 0x1000);
		break;
	case 1:
		memcpy(&nes_vram[0x1000], chr_rom_ptr + (chrrom * 0x1000),
		       0x1000);
		break;
	}
}

void nes_set_chr_page1k_by_addr(u16 addr, u8 chrrom)
{
	if (!chr_rom_ptr)
		return;

	if ((chrrom / 8) >= current_nes_file->chr_count) {
/*		fprintf(stdout, "TRIED TO GET CHR BANK %d -- DOESN'T EXIST\n", chrrom);*/
		chrrom %= (current_nes_file->chr_count * 8);
	}
	memcpy(&nes_vram[addr], chr_rom_ptr + (chrrom * 1024), 1024);
}

#ifdef TARGET_DC
void iopcode(u32 val)
{
	fprintf(stdout, "%04X:%02X\n", val, nes_read_byte(val));
}

void printstatus()
{
#if 1
	extern u8 _p, _x, _y, _a, _sp;
	extern u16 _pc;

	fprintf(stdout, "[PC:%04X] [@PC:%02X] %s\n",
		_pc, nes_read_byte(_pc), onames[nes_read_byte(_pc)]);
	fprintf(stdout, "[P:%02X] [X:%02X] [Y:%02X] [A:%02X] [SP:%02X]\n",
		_p, _x, _y, _a, _sp);
	fprintf(stdout, "[P:%c%c-%c%c%c%c%c] [@SP:%02X]\n",
		(_p & 0x80) ? 'N' : 'n',
		(_p & 0x40) ? 'V' : 'v',
		(_p & 0x10) ? 'B' : 'b',
		(_p & 0x08) ? 'D' : 'd',
		(_p & 0x04) ? 'I' : 'i',
		(_p & 0x02) ? 'Z' : 'z',
		(_p & 0x01) ? 'C' : 'c', nes_read_byte(0x0100 + _sp));
#endif
}

void errmsg(u32 val, u32 val2)
{
	extern u8 _p, _x, _y, _a, _sp;
	extern u16 _pc;

	fprintf(stdout, "Unknown opcode %02X [PC:%04X]\n", val, val2 - 1);
	fprintf(stdout, "[P:%02X] [X:%02X] [Y:%02X] [A:%02X] [SP:%02X]\n",
		_p, _x, _y, _a, _sp);
	fprintf(stdout, "REAL: [PC:%04X]\n", _pc - 1,
		nes_read_byte(_pc - 1));
}

u32 asm_cpu_interrupt()
{				/* called from the asm core */
	u32 ret = 0;

	if (next_line() == NMI_VECTOR) {
		if (PPUCNT1 & 0x80)
			ret = 1;	/* 1 == NMI */
		gleam_poll_events();
	}
	gleam_process_sound();
	return ret;
}
#endif
