#ifndef TARGET_DC
#include <stdio.h>
#include <stdlib.h>
#else
#include "dream.h"
#include "libc.h"
#include "dc.h"
#endif

#include "g6502.h"
#include "fileio.h"

nes_file *nes_file_open(u8 * filename)
{
	FILE *inf = (FILE *) NULL;
	nes_file *nf = (nes_file *) NULL;
	u32 filesize = 0;

	fprintf(stdout, "nes_file_open() ->\n");

	inf = fopen((char *) filename, "rb");
	if (!inf) {
		gleamerror("fopen(): unable to open %s\n", filename);
		return (nes_file *) NULL;
	}

	fseek(inf, 0, SEEK_END);
	filesize = ftell(inf);
	rewind(inf);

	fprintf(stdout, ":: Filesize: [%lu] bytes\n", filesize);

	nf = (nes_file *) malloc(sizeof(nes_file));
	fread((char *) nf, sizeof(nes_file) - sizeof(nf->data), 1, inf);

	nf->data =
	    (u8 *) malloc(filesize -
			  (sizeof(nes_file) - sizeof(nf->data)));
	fread((char *) nf->data,
	      filesize - (sizeof(nes_file) - sizeof(nf->data)), 1, inf);

	fclose(inf);

	fprintf(stdout, "<- nes_file_open();\n");
	return (nes_file *) nf;
}


void nes_file_details(nes_file * nf)
{
	u8 mapper = 0;

	fprintf(stdout, "nes_file_details() ->\n");
	fprintf(stdout, ":: Header: [");
	fwrite(nf->hdr, 3, 1, stdout);
	fprintf(stdout, "]\n");
	fprintf(stdout, ":: ID [0x%02X]\n", nf->id);
	fprintf(stdout, ":: PRG-COUNT: [%d] x 16kB\n", nf->prg_count);
	fprintf(stdout, ":: CHR-COUNT: [%d] x  8kB\n", nf->chr_count);

	switch (nf->ucb[0] & 0x01) {
	case 0:
		fprintf(stdout, ":: Mirroring: [Horizontal]\n");
		break;
	case 1:
		fprintf(stdout, ":: Mirroring: [Vertical]\n");
		break;
	}


	if (nf->ucb[0] & 0x02) {
		fprintf(stdout, ":: SRAM: [Enabled]\n");
	} else {
		fprintf(stdout, ":: SRAM: [Disabled]\n");
	}

	if (nf->ucb[0] & 0x04) {
		fprintf(stdout, ":: Trainer: [Present] x 512 bytes\n");
	} else {
		fprintf(stdout, ":: Trainer: [Not Present]\n");
	}

	if (nf->ucb[0] & 0x08) {
		fprintf(stdout, ":: Four-screen VRAM Layout: [Enabled]\n");
	} else {
		fprintf(stdout,
			":: Four-screen VRAM Layout: [Disabled]\n");
	}

	mapper = ((nf->ucb[0] & 0xF0) >> 4) | (nf->ucb[1] & 0xF0);
	fprintf(stdout, ":: Mapper: [%d]\n", mapper);

	fprintf(stdout, "<- nes_file_details();\n");

	return;
}
