#include <stdarg.h>

#ifndef TARGET_DC
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#else
#include "dream.h"
#include "libc.h"
#include "dc.h"
#endif

#include "gleam!.h"
#include "fileio.h"
#include "emulate.h"
#include "palette.h"

nes_file *current_nes_file = (nes_file *) NULL;
u8 debug_rom = 0;
char current_filename[128];

void sigint(int);
void gleam_graphics();

unsigned long time_start = 0;

#ifndef TARGET_DC
int main(int argc, char *argv[])
#else
int dc_main()
#endif
{
	u8 *filename;
#ifdef TARGET_DC
	int argc = 2;
	char *argv[] = { "gleam!-dc.srec", "                          " };
	dc_init();
#endif
	fprintf(stdout, "-> main()\n");
	time_start = time(NULL);
	fprintf(stdout, "Time started at %d\n", time_start);
	puts(VERSHDR "\n");

	signal(SIGINT, sigint);

	gleam_graphics();

#ifdef TARGET_DC
	dc_intro_screen();

	dc_gleam_option_screen(current_filename);
	while (1) {
		reset_colors();
#endif
		if (argc == 2) {
#ifdef TARGET_DC
			filename = current_filename;
#else
			filename = (u8 *) argv[1];
#endif
			current_nes_file = nes_file_open(filename);
			if (current_nes_file) {
				nes_file_details(current_nes_file);

				nes_emulate(current_nes_file);
			}
		} else if (argc == 3) {
			filename = (u8 *) argv[1];
			current_nes_file = nes_file_open(filename);
			debug_rom = 1;
			if (current_nes_file) {
				nes_file_details(current_nes_file);
				nes_emulate(current_nes_file);
			}
		}
#ifdef TARGET_DC
		malloc(0);
	}			/* game loop */
#endif

	fprintf(stdout, "<- main();\n");
#ifndef TARGET_DC
	sigint(0);
#endif
	/*exit(1); /* better compatability */
	return;
}


int gleamerror(char *format, ...)
{
	char buffer[2048];
	va_list args;

	va_start(args, format);
	vsprintf(buffer, format, args);
	va_end(args);

	puts(buffer);
	return 0;
}

void sigint(int signo)
{
	FILE *fp;
	fprintf(stdout, "-> sigint()\n");
	fp = fopen("vram.mem", "w");
	fwrite(nes_vram, 0x4000, 1, fp);
	fclose(fp);

	fp = fopen("ram.mem", "w");
	fwrite(nes_ram, 0x800, 1, fp);
	fclose(fp);

	fp = fopen("sram.mem", "w");
	fwrite(nes_sram, 0x2000, 1, fp);
	fclose(fp);

	fp = fopen("regs.mem", "w");
	fwrite((void *) nes_regs, 8, 2, fp);
	fclose(fp);
	fprintf(stdout, "<- sigint();\n");
	exit(1);
}



void gleam_graphics()
{
#ifndef TARGET_DC
	gleam_init_gl(NULL);
	gleam_create_window_gl("gleam!", NES_SCREEN_WIDTH,
			       NES_SCREEN_HEIGHT);

#else				/* TARGET_DC */
	/*dc_vid_init(dc_vid_check_cable(), PM_RGB555); */
	dc_init_video2(dc_vid_check_cable(), PM_RGB555, 0);
	dc_maple_init();
	dc_vid_clear(0, 0, 0);
	reset_colors();

#endif
}

void reset_colors()
{
	int i;

	for (i = 0; i < 256; i++) {
		set_color(i, palette[i].red, palette[i].green,
			  palette[i].blue);
	}
	set_color(73, 0, 0, 0);	/* black */
	set_color(74, 0x1f, 0x1f, 0x1f);	/* white */
	set_color(75, 0, 0, 0x1f);	/* dark blue */
	set_color(76, 0, 0, 0x07);	/* light blue */
	set_color(77, 0x1f, 0, 0);	/* dark red */
}
