#ifndef TARGET_DC

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#include <X11/Xlib.h>
#include <X11/keysym.h>

#include <GL/gl.h>
#include <GL/glx.h>

#include "gleam!.h"
#include "emulate.h"

int attrib[] = {
	GLX_RGBA,
	GLX_RED_SIZE, 5,
	GLX_GREEN_SIZE, 5,
	GLX_BLUE_SIZE, 5,
	GLX_DOUBLEBUFFER,
	None
};

Display *dpy;
XVisualInfo *vi;
GLXContext ctx;
Window win;
XSetWindowAttributes attr;
Colormap cmap;
Atom wmDeleteWindow;
XSizeHints *sizehints;

GLuint texture;

#define GFX_BITDEPTH (16 / 8)

#define RGBA
#define TEXTURE


unsigned char cpal[256 * 4];

static unsigned char *gfx_buffer = (unsigned char *) NULL;
static unsigned long
 gleam_window_width = 256, gleam_window_height = 240;

int gleam_init_gl(char *display)
{
	dpy = XOpenDisplay(display);

	if (dpy == NULL) {
		fprintf(stderr, "Failed to open X connect to %s\n",
			XDisplayName(display));
		return 0;
	}

	if (glXQueryExtension(dpy, NULL, NULL) == False) {
		fprintf(stderr, "X server %s does not support GLX\n",
			XDisplayName(NULL));
		exit(EXIT_FAILURE);
	}

	vi = glXChooseVisual(dpy, DefaultScreen(dpy), attrib);

	if (vi == NULL) {
		fprintf(stderr, "No usable GL visual found\n");
		exit(EXIT_FAILURE);
	}

	ctx = glXCreateContext(dpy, vi, NULL, True);

	if (ctx == NULL) {
		fprintf(stderr, "glx context create failed\n");
		exit(EXIT_FAILURE);
	}

	return GFX_BITDEPTH;
}

int gleam_close_gl()
{
}

int gleam_create_window_gl(char *title, int width, int height)
{
	int major, minor;

	attr.background_pixel = BlackPixel(dpy, DefaultScreen(dpy));
	attr.event_mask =
	    KeyPressMask | KeyReleaseMask | ExposureMask |
	    StructureNotifyMask;

	cmap =
	    XCreateColormap(dpy, RootWindow(dpy, DefaultScreen(dpy)),
			    vi->visual, AllocNone);
	attr.colormap = cmap;

	win = XCreateWindow(dpy, RootWindow(dpy, DefaultScreen(dpy)), 0, 0,
			    width, height, 0, vi->depth, InputOutput,
			    vi->visual,
			    CWEventMask | CWColormap | CWBackPixel, &attr);

	sizehints = XAllocSizeHints();
	if (sizehints) {
		sizehints->flags = 0;
	}

	XSetWMProperties(dpy, win, NULL, NULL, NULL, 0, sizehints, None,
			 None);
	/* TODO: use XWMHints struct to set icon etc */

	if (sizehints)
		XFree(sizehints);

	XStoreName(dpy, win, title);
	XSetIconName(dpy, win, title);

	wmDeleteWindow = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
	XSetWMProtocols(dpy, win, &wmDeleteWindow, 1);

	glXMakeCurrent(dpy, win, ctx);

	if (glXQueryVersion(dpy, &major, &minor) == False) {
		fprintf(stderr, "glXQueryVersion returned False?\n");
		exit(EXIT_FAILURE);
	} else {
		printf("GLX Version %d.%d\n", major, minor);
		/* glXGetClientString / glXQueryServerString / glXQueryExtensionsString */
	}

	printf("GL_VENDOR: %s\n", glGetString(GL_VENDOR));
	printf("GL_RENDERER: %s\n", glGetString(GL_RENDERER));
	printf("GL_VERSION: %s\n", glGetString(GL_VERSION));
	printf("GL_EXTENSIONS: %s\n", glGetString(GL_EXTENSIONS));

	XMapWindow(dpy, win);

	XFlush(dpy);

	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();
	glMatrixMode(GL_COLOR);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);

	glShadeModel(GL_FLAT);
	glDisable(GL_DITHER);

	glClear(GL_COLOR_BUFFER_BIT);

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	memset(cpal, 255, sizeof(cpal));
	reset_colors();

#ifdef TEXTURE
	glEnable(GL_TEXTURE_2D);

	glNewList(1, GL_COMPILE);

	glBegin(GL_QUADS);
	glTexCoord2f(256.0f / 256.0f, 0.0f);
	glVertex2f(1.0f, 1.0f);
	glTexCoord2f(256.0f / 256.0f, 240.0f / 256.0f);
	glVertex2f(1.0f, -1.0f);
	glTexCoord2f(0.0f, 240.0f / 256.0f);
	glVertex2f(-1.0f, -1.0f);
	glTexCoord2f(0.0f, 0.0f);
	glVertex2f(-1.0f, 1.0f);
	glEnd();

	glEndList();

	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

#ifdef RGBA
	gfx_buffer = (unsigned char *) malloc(256 * 256 * 4);
	memset(gfx_buffer, 255, 256 * 256 * 4);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 256, 256, 0, GL_RGBA,
		     GL_UNSIGNED_BYTE, gfx_buffer);
	free(gfx_buffer);
	gfx_buffer = (unsigned char *) malloc(256 * 240 * 4);
	memset(gfx_buffer, 255, 256 * 240 * 4);
#else
	gfx_buffer = (unsigned char *) malloc(256 * 256 * 3);
	memset(gfx_buffer, 255, 256 * 256 * 3);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 256, 256, 0, GL_RGB,
		     GL_UNSIGNED_BYTE, gfx_buffer);
	free(gfx_buffer);
	gfx_buffer = (unsigned char *) malloc(256 * 240 * 3);
	memset(gfx_buffer, 255, 256 * 240 * 3);
#endif

#else				/* TEXTURE */

	glPixelZoom(1.0f, -1.0f);
	glRasterPos2f(-1.0f, 1.0f);

#ifdef RGBA
	gfx_buffer = (unsigned char *) malloc(256 * 240 * 4);
	memset(gfx_buffer, 255, 256 * 240 * 4);
#else
	gfx_buffer = (unsigned char *) malloc(256 * 240 * 3);
#endif

#endif				/* TEXTURE */


}

static void handler_keypress(KeySym key)
{
	extern float vblmax;
	switch (key) {
	case XK_Escape:
		printf("\nmax fps = %f\n", vblmax);
		exit(1);
		break;
	case XK_Return:
		JOYPAD1 |= 0x08;
		break;
	case XK_Tab:
		JOYPAD1 |= 0x04;
		break;
	case XK_Control_L:
		JOYPAD1 |= 0x02;
		break;
	case XK_Alt_L:
		JOYPAD1 |= 0x01;
		break;
	case XK_Left:
		JOYPAD1 |= 0x40;
		break;
	case XK_Right:
		JOYPAD1 |= 0x80;
		break;
	case XK_Up:
		JOYPAD1 |= 0x10;
		break;
	case XK_Down:
		JOYPAD1 |= 0x20;
		break;
	}
}

static void handler_keyunpress(KeySym key)
{
	switch (key) {
	case XK_Return:
		JOYPAD1 &= ~0x08;
		break;
	case XK_Tab:
		JOYPAD1 &= ~0x04;
		break;
	case XK_Control_L:
		JOYPAD1 &= ~0x02;
		break;
	case XK_Alt_L:
		JOYPAD1 &= ~0x01;
		break;
	case XK_Left:
		JOYPAD1 &= ~0x40;
		break;
	case XK_Right:
		JOYPAD1 &= ~0x80;
		break;
	case XK_Up:
		JOYPAD1 &= ~0x10;
		break;
	case XK_Down:
		JOYPAD1 &= ~0x20;
		break;
	}
}

int testit()
{
	static u32 y = 0;
	y++;
	if ((y & 15) == 7)
		return 1;
}

int gleam_poll_events()
{
	XEvent event;

	while (XPending(dpy)) {
		XNextEvent(dpy, &event);

		/* TODO: watch focus and clear key status when focus changes */
		switch (event.type) {
		case KeyPress:
			handler_keypress(XKeycodeToKeysym
					 (dpy, event.xkey.keycode, 0));
			break;
		case KeyRelease:
			handler_keyunpress(XKeycodeToKeysym
					   (dpy, event.xkey.keycode, 0));
			break;
		case ConfigureNotify:
			glViewport(0, 0, event.xconfigure.width,
				   event.xconfigure.height);
#ifndef TEXTURE
			glPixelZoom((double) event.xconfigure.width /
				    (double) 160.0,
				    (double) -event.xconfigure.height /
				    (double) 144.0);
			glRasterPos2f(-1.0f, 1.0f);
#endif
		case ClientMessage:
			if (event.xclient.data.l[0] == wmDeleteWindow)
				handler_keypress(XK_Escape);	/* HACK */
			break;
		default:
			/*printf("What is this: %d\n", event.type); */
			break;
		}
		XFlush(dpy);
	}

	return 1;
}


void vram_blit()
{
#if 0
	int error = glGetError();
	if (error != GL_NO_ERROR) {
		fprintf(stderr, "error -> %d\n", error);
		exit(EXIT_FAILURE);
	}
#endif
#ifdef TEXTURE
	/* glClear(GL_COLOR_BUFFER_BIT); */
#ifdef RGBA
	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 256, 240, GL_RGBA,
			GL_UNSIGNED_BYTE, gfx_buffer);
#else
	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 256, 240, GL_RGB,
			GL_UNSIGNED_BYTE, gfx_buffer);
#endif
	glCallList(1);

#else				/* TEXTURE */

#ifdef RGBA
	glDrawPixels(gleam_window_width, gleam_window_height, GL_RGBA,
		     GL_UNSIGNED_BYTE, gfx_buffer);
#else				/* RGBA */
	glDrawPixels(gleam_window_width, gleam_window_height, GL_RGB,
		     GL_UNSIGNED_BYTE, gfx_buffer);
#endif				/* RGBA */

#endif				/* TEXTURE */

	glXSwapBuffers(dpy, win);

}

void set_color(int num, unsigned short int red, unsigned short int green,
	       unsigned short int blue)
{
	cpal[(num << 2) + 0] = red * 4;
	cpal[(num << 2) + 1] = green * 4;
	cpal[(num << 2) + 2] = blue * 4;
}

void vram_plot(int x, int y, int color)
{
	unsigned long *pald, *ptrd;
	unsigned char *ptr;
	unsigned char r, g, b;

#ifdef RGBA
	pald = (unsigned long *) &cpal[color << 2];
	ptrd = (unsigned long *) gfx_buffer + y * (gleam_window_width) + x;
	*ptrd = *pald;

#else
	ptr = gfx_buffer + (y * (gleam_window_width * 3)) + (x * 3);
	*ptr = cpal[(color << 2) + 0];
	ptr++;
	*ptr = cpal[(color << 2) + 1];
	ptr++;
	*ptr = cpal[(color << 2) + 2];
#endif
}


inline void vram_plot_line(int y, u8 * color)
{
	unsigned long *pald, *ptrd, x;
	unsigned char *ptr;
	unsigned char r, g, b;

#ifdef RGBA
	ptrd = (unsigned long *) gfx_buffer + y * (gleam_window_width);
	for (x = 0; x < 256; x++) {
		pald = (unsigned long *) &cpal[color[0] << 2];
		*ptrd = *pald;
		color++;
		ptrd++;
	}
#else
	ptr = gfx_buffer + (y * (gleam_window_width * 3));
	for (x = 0; x < 256 * 3; x++) {
		int c = *color++;
		*ptr = cpal[(c << 2) + 0];
		ptr++;
		*ptr = cpal[(c << 2) + 1];
		ptr++;
		*ptr = cpal[(c << 2) + 2];
		ptr++;
	}
#endif
}






#endif	/* TARGET_DC */
