#ifndef TARGET_DC
#include <stdio.h>
#include <stdlib.h>
#else
#include "dream.h"
#include "libc.h"
#include "dc.h"
#endif


#include "gleam!.h"
#include "types.h"
#include "g6502.h"
#include "mem.h"
#include "fileio.h"
#include "mappers.h"
#include "mappers/m0.h"
#include "mappers/m1.h"
#include "mappers/m2.h"
#include "mappers/m3.h"
#include "mappers/m4.h"
#include "mappers/m7.h"
#include "mappers/m8.h"

struct _mapper nes_mappers[] = {
	{0, (u8 *) "No Mapper", m0_write, m0_read, m0_init},
#if 1				/* unused right now */
	{1, (u8 *) "MMC1", m1_write, m1_read, m1_init},
#endif
	{2, (u8 *) "UNROM", m2_write, m2_read, m2_init},
	{3, (u8 *) "CNROM", m3_write, m3_read, m3_init},
	{4, (u8 *) "MMC3", m4_write, m4_read, m4_init},
	{7, (u8 *) "AOROM", m7_write, m7_read, m7_init},
	{8, (u8 *) "FFE F3xxx", m8_write, m8_read, m8_init},
	{65, (u8 *) "MMC1", m1_write, m1_read, m1_init},
	{0, (u8 *) NULL, 0, 0, 0},
};


struct _mapper *current_mapper = (struct _mapper *) NULL;

void nes_mapper_init(nes_file * nf)
{
	u8 mapperno = 0;
	u8 cnt = 0;

	mapperno = ((nf->ucb[0] & 0xF0) >> 4) | (nf->ucb[1] & 0xF0);
	mapperno &= 0x0F;

	for (cnt = 0; nes_mappers[cnt].name; cnt++) {
		if (nes_mappers[cnt].number == mapperno)
			break;
	}

	if (!nes_mappers[cnt].name) {
		fprintf(stdout,
			":: Could not find mapper %d, defaulting to 0\n",
			mapperno);
		cnt = 0;
	}
	current_mapper = &nes_mappers[cnt];
	current_mapper->init_func(nf);

}

void nes_mapper_write(u16 address, u8 value)
{
	current_mapper->write_func(address, value);
}

u8 nes_mapper_read(u16 address)
{
	return current_mapper->read_func(address);
}
