#ifndef TARGET_DC
#include <stdio.h>
#include <stdlib.h>
#else
#include "dream.h"
#include "libc.h"
#include "dc.h"
#endif

#include "gleam!.h"
#include "emulate.h"
#include "m1.h"

u8 registers[4] = { 0x0C, 0, 0, 0 };
u8 regbitcnt[4] = { 0, 0, 0, 0 };

u8 cart_type = 0; /* 0 = < 512 | 1 == 512kb | 2 == 1024kb */

u8 m1_read(u16 address)
{
        switch(address & 0xF000) {
                case 0x8000:
                case 0x9000:
                        return prg_banks[0][address & 0x1FFF];
                case 0xA000:
                case 0xB000:
                        return prg_banks[1][address & 0x1FFF];
                case 0xC000:
                case 0xD000:
                        return prg_banks[2][address & 0x1FFF];
                default:
                        return prg_banks[3][address & 0x1fff];
        }
}

void m1_write(u16 address, u8 value)
{
	u8 bit;

	bit = value & 0x01;

	switch(address & 0xE000) {
		case 0x8000: /* Register 0 */
			if(value & 0x80) {
				regbitcnt[0] = 0;
				registers[0] = 0x0C;
				current_nes_file->ucb[0] |= 0x01;
				nes_single_screen_mirroring = 0;
				break;
			}
			registers[0] &= ~(1 << regbitcnt[0]);
			registers[0] |= bit << regbitcnt[0];
			regbitcnt[0]++;

			if(regbitcnt[0] == 5) { /* handle register */
				if(!(registers[0] & 0x01)) {
					current_nes_file->ucb[0] |= 0x01;
				printf("vertical selected\n");
				} else {
					current_nes_file->ucb[0] &= ~0x01;
				printf("horizontal selected\n");
				}
				
				if(registers[0] & 0x02) {
					nes_single_screen_mirroring = 0;
				} else {
					nes_single_screen_mirroring = 1;
				}
				if((registers[0] & 0xC0) == 0xC0) {
					nes_set_rom_page(1, current_nes_file->prg_count - 1);
				} else if((registers[0] & 0xC0) == 0x80) {
					nes_set_rom_page(0, 0);
				}
				regbitcnt[0] = 0;
			}
			
			break;
		case 0xA000: /* Register 1 */
			if(value & 0x80) {
				regbitcnt[1] = 0;
				break;
			}
			if(regbitcnt[1] == 0)
				registers[1] = 0;
			registers[1] |= bit << regbitcnt[1];
			regbitcnt[1]++;
			
			if(regbitcnt[1] == 5) { /* handle register */
				u8 chrbank = registers[1] & 0x0F;

				if(registers[0] & 0x10) {
					nes_set_chr_page(0, chrbank);
				} else {
					nes_set_chr_page8k(chrbank>>1);
				}
				
				regbitcnt[1] = 0;
			}
			break;
		case 0xC000: /* Register 2 */
			if(value & 0x80) {
				regbitcnt[2] = 0;
				break;
			}
			if(regbitcnt[2] == 0)
				registers[2] = 0;
			
			registers[2] |= bit << regbitcnt[2];
			regbitcnt[2]++;

			if(regbitcnt[2] == 5) { /* handle register */
				u8 chrbank = registers[2] & 0x0F;

				if(registers[0] & 0x10) {
					nes_set_chr_page(1, chrbank);
				}
				regbitcnt[2] = 0;
			}
			break;
		case 0xE000: /* Register 3 */
			if(value & 0x80) {
				regbitcnt[3] = 0;
				break;
			}
			if(regbitcnt[3] == 0)
				registers[3] = 0;
			registers[3] |= bit << regbitcnt[3];
			regbitcnt[3]++;
			
			if(regbitcnt[3] == 5) { /* handle register */
				u8 prgbank = registers[3] & 0x1F;
				
				switch(cart_type) {
					case 1:
						prgbank += (registers[1] & 0x10);
						nes_set_rom_page(0, prgbank);
						break;
					case 2:
						if(registers[0] & 0x10) {
							u8 val = 0;
							val |= (registers[1] & 0x10) >> 4;
							val |= (registers[2] & 0x10) >> 3;
							prgbank += (16 * val);
						} else {
							if(registers[1] & 0x10)
								prgbank += (16 * 2);
						}
						nes_set_rom_page(0, prgbank);
						break;
					default:
						if(registers[0] & 0x08) {
							if(registers[0] & 0x04) {
								nes_set_rom_page(0, prgbank);
								nes_set_rom_page(1, current_nes_file->prg_count - 1);
							} else {
								nes_set_rom_page(0, 0);
								nes_set_rom_page(1, prgbank);
							}
						} else {
							prgbank >>= 1;
							nes_set_rom_page32k(prgbank);
						}
						break;
				}
				regbitcnt[3] = 0;
			}
			break;
	}
}

void m1_init(nes_file *nf)
{
	nes_set_rom_page(0, 0);
	
	if(nf->prg_count <= 16)
		nes_set_rom_page(1, nf->prg_count - 1);
	else if(nf->prg_count >= 32)
		nes_set_rom_page(1, 15);
		
	if(nf->prg_count == 32)
		cart_type = 1;
	else if(nf->prg_count == 64)
		cart_type = 2;
	else
		cart_type = 0;
}
