#ifndef TARGET_DC
#include <stdio.h>
#include <stdlib.h>
#else
#include "dream.h"
#include "libc.h"
#include "dc.h"
#endif

#include "gleam!.h"
#include "emulate.h"
#include "m4.h"

u8 registers[7];

u8 irq_count_down = 0;
u8 irq_latch_value = 0;
u8 irq_control = 0;

u8 m4_read(u16 address)
{
        switch(address & 0xF000) {
                case 0x8000:
                case 0x9000:
                        return prg_banks[0][address & 0x1FFF];
                case 0xA000:
                case 0xB000:
                        return prg_banks[1][address & 0x1FFF];
                case 0xC000:
                case 0xD000:
                        return prg_banks[2][address & 0x1FFF];
                default:
                        return prg_banks[3][address & 0x1fff];
        }
}

void m4_write(u16 address, u8 value)
{
	u16 addr;
	
	switch(address) {
		case 0x8000:
			registers[0] = value & 0xC7;
			break;
		case 0x8001:
/*			value &= 0x7F;*/
			switch(registers[0] & 0x07) {
				case 0:
					value &= 0xFE;
					addr = 0x0000;
					if(registers[0] & 0x80)
						addr ^= 0x1000;
					nes_set_chr_page1k_by_addr(addr, value);
					nes_set_chr_page1k_by_addr(addr + 1024, value + 1);
					break;
				case 1:
					value &= 0xFE;
					addr = 0x0800;
					if(registers[0] & 0x80)
						addr ^= 0x1000;
					nes_set_chr_page1k_by_addr(addr, value);
					nes_set_chr_page1k_by_addr(addr + 1024, value + 1);
					break;
				case 2:
					addr = 0x1000;
					if(registers[0] & 0x80)
						addr ^= 0x1000;
					nes_set_chr_page1k_by_addr(addr, value);
					break;
				case 3:
					addr = 0x1400;
					if(registers[0] & 0x80)
						addr ^= 0x1000;
					nes_set_chr_page1k_by_addr(addr, value);
					break;
				
				case 4:
					addr = 0x1800;
					if(registers[0] & 0x80)
						addr ^= 0x1000;
					nes_set_chr_page1k_by_addr(addr, value);
					break;
				
				case 5:
					addr = 0x1C00;
					if(registers[0] & 0x80)
						addr ^= 0x1000;
					nes_set_chr_page1k_by_addr(addr, value);
					break;
				case 6:
					if(registers[0] & 0x40) {
						nes_set_rom_page8k(1, value);
					} else {
						nes_set_rom_page8k(0, value);
					}
					break;
				case 7:
					if(registers[0] & 0x40) {
						nes_set_rom_page8k(2, value);
					} else {
						nes_set_rom_page8k(1, value);
					}
					break;
			}
			break;
		case 0xA000:
			if(!(value & 0x01)) {
				current_nes_file->ucb[0] |= 0x01;
			} else {  
				current_nes_file->ucb[0] &= ~0x01;
			}
			break;
		case 0xC000:
			irq_count_down = value;
			break;
		case 0xC001:
			irq_latch_value = value;
			break;
		case 0xE000:
			irq_control = 0;
			irq_count_down = irq_latch_value;
			irq_latch_value = 0;
			break;
		case 0xE001:
			irq_control = 1;
			break;
	}	
}

void m4_init(nes_file *nf)
{
	nes_set_rom_page8k(0, 0);
	nes_set_rom_page8k(1, 1);
	nes_set_rom_page(1, nf->prg_count - 1);	
	
	if(nf->chr_count) {
		nes_set_chr_page8k(0);
	}
}
