#ifndef TARGET_DC
#include <stdio.h>
#include <stdlib.h>
#else
#include "dream.h"
#include "libc.h"
#include "dc.h"
#endif

#include "gleam!.h"
#include "fileio.h"
#include "emulate.h"
#include "mem.h"
#include "mappers.h"
#include "ppu.h"

static u8 garbage_read = 1;
static u8 dbl_regs[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
static u8 joy_bit[2] = { 0, 0 };

inline void nes_mem_reset()
{
	int i;
	for (i = 0; i < 8; i++)
		dbl_regs[i] = 0;
	joy_bit[0] = 0;
	joy_bit[1] = 0;
	garbage_read = 1;
}

inline u8 vram_read(u16 addr)
{

	return nes_vram[addr];
}

inline u8 vram_write(u16 addr, u8 value)
{
/*fprintf(stdout, "VRAMWRITE: %02X -> %04X\n", value, addr);*/
	if ((addr & 0xff00) == 0x3f00) {
		addr &= 0xff1f;
		switch (addr & 0x00ff) {
		case 0x10:
			nes_vram[0x3f00] =
			    nes_vram[0x3f04] =
			    nes_vram[0x3f08] =
			    nes_vram[0x3f0c] =
			    nes_vram[0x3f10] =
			    nes_vram[0x3f14] =
			    nes_vram[0x3f18] = nes_vram[0x3f1c] = value;
			break;
		default:
			nes_vram[addr] = value;
		}
	} else {
		nes_vram[addr] = value;
	}
}

#if 0
inline u8 nes_read_byte0(u16);
inline u8 nes_read_byte(u16 address)
{
	u8 val = nes_read_byte0(address);
	fprintf(stdout, "MEMREAD: %04X[%02X]\n", address, val);
	return val;
}

inline u8 nes_read_byte0(u16 address)
#else
inline u8 nes_read_byte(u16 address)
#endif
{
/*	extern u16 _pc;
	if(address == _pc)
		fprintf(stdout, "MEMREAD %04X\n", address);
*/
	switch (address & 0xF000) {
	case 0x8000:
	case 0x9000:
		return prg_banks[0][address & 0x1FFF];
	case 0xA000:
	case 0xB000:
		return prg_banks[1][address & 0x1FFF];
	case 0xC000:
	case 0xD000:
		return prg_banks[2][address & 0x1FFF];
	case 0xE000:
	case 0xF000:
		return prg_banks[3][address & 0x1FFF];

	case 0x0000:
	case 0x1000:
		return nes_ram[address & 0x07FF];
	case 0x2000:
	case 0x3000:
		{
			u8 reg = address & 7;
			switch (reg) {
			case 0:
			case 1:
			case 5:
			case 6:
				return nes_regs[reg];
			case 3:
				return nes_regs[3];
			case 2:
				dbl_regs[6] = dbl_regs[5] = 0;
				return nes_ppu_get_status();
			case 4:
				return nes_sprram[nes_regs[3]++];
			case 7:
				{
					u8 val;
					VRAMADR &= 0x3FFF;
					val = vram_read(VRAMADR);
					if (garbage_read) {
						garbage_read = 0;
						return val;
					}
					if (PPUCNT1 & PPUINC) {
						VRAMADR += 0x20;
					} else {
						VRAMADR++;
					}
					return val;
				}
			}
			break;
		}
	case 0x4000:
		{
			u8 ret = 0;
			switch (address) {
			case 0x4015:
				ret = gleam_sound_read(0x15);
				break;
			case 0x4016:
				ret = (JOYPAD1 & joy_bit[0]) ? 1 : 0;
				joy_bit[0] <<= 1;
				if (joy_bit[0] == 0)
					joy_bit[0] = 1;
				break;
			case 0x4017:
				ret = (JOYPAD2 & joy_bit[1]) ? 1 : 0;
				joy_bit[1] <<= 1;
				if (joy_bit[1] == 0)
					joy_bit[1] = 1;
				break;
			default:
				break;
			}
			return ret;
			break;
		}
	case 0x5000:
		{
			/* expansion shit */
			return 0;
		}
	case 0x6000:
	case 0x7000:
		/* sram */
		return nes_sram[address - 0x6000];
	}

	return 0;
}

void nes_write_byte(u16 address, u8 data)
{
/*fprintf(stdout, "MEMWRITE: %02X -> %04X\n", data, address);*/
	switch (address & 0xF000) {
	case 0x0000:
	case 0x1000:
		nes_ram[address & 0x07FF] = data;
		break;
	case 0x2000:
	case 0x3000:
		{
			u8 reg = address & 0x07;

			switch (reg) {
			case 0:	/* 2000 and 2001 */
			case 1:
				nes_regs[reg] = data;
				break;
			case 2:	/* 2002 */
				nes_regs[reg] = data;
				break;
			case 3:	/* 2003 */
				nes_regs[reg] = data;
				break;
			case 4:	/* 2004 */
				nes_sprram[nes_regs[3]++] = data;
				break;
			case 5:	/* 2005 */
				if (dbl_regs[reg]) {
					nes_regs[reg] =
					    (nes_regs[reg] & 0xFF00) |
					    data;
				} else {
					nes_regs[reg] =
					    (data << 8) | (nes_regs[reg] &
							   0x00FF);
				}
				dbl_regs[reg] ^= 1;
				break;
			case 6:	/* 2006 */
				if (dbl_regs[reg]) {
					nes_regs[reg] =
					    (nes_regs[reg] & 0xFF00) |
					    data;
				} else {
					nes_regs[reg] =
					    (data << 8) | (nes_regs[reg] &
							   0x00FF);
				}
				dbl_regs[reg] ^= 1;
				garbage_read = 1;
				break;
			case 7:	/* 2007 */
				if (nes_ppu_get_status() & 0x10) {
					printf("not writing..nana\n");
					break;
				}
				VRAMADR &= 0x3FFF;
				vram_write(VRAMADR, data);


				if ((VRAMADR & 0x3000) == 0x2000) {
					if (nes_single_screen_mirroring) {
						vram_write(VRAMADR ^ 0x400,
							   data);
						vram_write(VRAMADR ^ 0x800,
							   data);
						vram_write((VRAMADR ^
							    0x400) ^ 0x800,
							   data);
					} else {
						if ((current_nes_file->
						     ucb[0] & 0x01) == 0) {
							vram_write(VRAMADR
								   ^ 0x400,
								   data);
						} else {
							vram_write(VRAMADR
								   ^ 0x800,
								   data);
						}
					}
				}

				if (PPUCNT1 & PPUINC) {
					VRAMADR += 0x20;
				} else {
					VRAMADR++;
				}
				break;
			}
			break;
		}
	case 0x4000:
		{
			if ((address & 0x00FF) < 0x14) {
				gleam_sound_write(address & 0xFF, data);
				break;
			}
			switch (address) {
			case 0x4015:
				gleam_sound_write(address & 0xFF, data);
				break;
			case 0x4014:
				{
					u16 j, a = 0x100 * data;
					for (j = 0; j < 256; j++)
						nes_sprram[j] =
						    nes_read_byte(a + j);
					break;
				}
			case 0x4016:
				if (data == 1)
					joy_bit[0] = joy_bit[1] = 1;
				break;
			case 0x4017:
				/*      if(data == 1)
				   joy_bit[1] = 1; */
				break;
			default:
				break;
			}
			break;
		}
	case 0x5000:
		/* expansion shit */
		break;
	case 0x6000:
	case 0x7000:
		/* sram */
		nes_sram[address - 0x6000] = data;
		break;
	default:
		nes_mapper_write(address, data);
	}

}



u16 nes_read_word(u16 address)
{
	return ((nes_read_byte(address)) |
		(nes_read_byte(address + 1) << 8));
}
