#ifndef TARGET_DC
#include <stdio.h>
#include <stdlib.h>
#else
#include "dream.h"
#include "libc.h"
#include "dc.h"
#endif

#include "gleam!.h"
#include "g6502.h"
#include "types.h"
#include "emulate.h"

#ifndef TARGET_DC
#	include "glx.h"
#endif


const int SCANLINE_MAX = 240;

u16 current_line = 0;
u8 sprite_hit = 0;
u8 in_vblank = 0;
u8 sprite_count = 0;
u8 dsprites[512];
u8 dprior[512];
u8 dbackground[512];

u8 nes_get_color_pixel(u8 x, u8 y);
u8 nes_get_color_sprite(u8 x, u8 y);

inline void draw_line(u8 line);
inline void nes_bg_line(u8);

u16 vblcount = 0, total_vblcount = 0;
int vblmax = 0;

extern unsigned long time_start;
extern u32 option_frameskip;

/* #define DONTSCROLL /* */
int otherscreens = 0;

u8 nes_ppu_get_status()
{
	u8 val = 0;

	if (in_vblank)
		val |= 0x80;
	if (sprite_hit)
		val |= 0x40;
	if (sprite_count > 8)
		val |= 0x20;
/*	if(current_line < SCANLINE_MAX)
		val |= 0x10;	/* ignore writes when not in vblank */

	in_vblank = 0;

	return val;
}


u16 next_line()
{
	u16 ret = 0;

	if (current_line >= SCANLINE_MAX) {
		if (current_line == (SCANLINE_MAX + 8)) {
			ret = NMI_VECTOR;
		} else if (current_line == SCANLINE_MAX) {
			in_vblank = 1;

			if (vblcount == (option_frameskip + 1)) {
				vram_blit();
				vblcount = 0;
			}
		}
		if (current_line == 256) {
			int tnow = time(NULL);
			current_line = 0xFFFF;
			sprite_hit = in_vblank = 0;
			vblcount++, total_vblcount++;

			if ((total_vblcount % 30) == 0) {
				int val;
				if ((tnow - time_start) != 0) {
					val =
					    (total_vblcount) /
					    ((tnow - time_start));
					if (val > vblmax
					    && (tnow != time_start))
						vblmax = val;
					fprintf(stdout, "%d fps | %d\r",
						val, vblmax);
					fflush(stdout);
				}
			}

		}

	} else {
		if (current_line == (nes_sprram[0] + 8))
			sprite_hit = 1;

		if (option_frameskip == 0
		    || vblcount == (option_frameskip + 1))
			draw_line(current_line);
	}
	current_line++;
	return ret;
}


inline void draw_line(u8 line)
{
	u16 x, y = line;
	u8 color;

	if (line >= SCANLINE_MAX)
		return;

	memset(dprior, 0, 512);
	memset(dsprites, 0, 512);
	if ((PPUCNT2 & 0x10) && otherscreens == 0)
		nes_get_color_sprite(0, y);

	/*memset(dbackground, 0, 512); */
	if (PPUCNT2 & 0x08)
		nes_bg_line(y);	/*nes_get_color_pixel_line(y); */
#if 0
	for (x = 0; x < 256; x++) {
		if (!(color = dsprites[x]));	/*color = dbackground[x]; */
		vram_plot(x, y, color);
	}
#else
	vram_plot_line(y, dbackground);
#endif
	return;
}

u8 nes_get_color_sprite(u8 x, u8 y)
{
	u8 i;
	u8 *sprptr = nes_sprram + 0xFF;
	u8 i8x16;
	u8 sprh = 8;

	/* Sprite variables */
	u8 sx, sy, st, sa;
	u8 sx_offset, sy_offset;
	u8 b0, b1, c23, ccolor, hflip, vflip, priority;
	u8 temp;
	u8 *tileptr;

	if ((PPUCNT1 & 0x20)) {
		i8x16 = 1;
		sprh += 8;
	} else
		i8x16 = 0;

	for (i = 0; i < 64; i++) {
		sx = *sprptr--;
		sa = *sprptr--;
		st = *sprptr--;
		sy = *sprptr-- + 1;

		/* if this pixel (x,y) is not part of this sprite (sx,sy) */
		if (y < sy)
			continue;
		if (y >= (sy + sprh))
			continue;

		sx_offset = 0;	/* do the entire 8 pixels of the sprite */
		sy_offset = (y - sy);

		hflip = (sa & 0x40) ? 1 : 0;
		vflip = (sa & 0x80) ? 1 : 0;
		priority = (sa & 0x20) ? 1 : 0;

		if (i8x16) {
			if (st & 0x01)
				tileptr = &nes_vram[0x1000];
			else
				tileptr = nes_vram;
		} else {
			if ((PPUCNT1 & 0x08))
				tileptr = &nes_vram[0x1000];
			else
				tileptr = nes_vram;
		}


		if (i8x16) {
			if (st & 0x01) {
				tileptr = &nes_vram[0x1000];
				st &= 0xFE;
			}

			if (sy_offset >= 8) {
				if (vflip) {
					sy_offset -= 8;
					sy_offset = 7 - sy_offset;
				} else {
					st++;
					sy_offset -= 8;
				}
			} else {
				if (vflip) {
					st++;
					sy_offset = 7 - sy_offset;
				}
			}
		} else {
			if (vflip)
				sy_offset = 7 - sy_offset;
		}

		tileptr += (st << 4) + sy_offset;
		c23 = sa & 0x03;

		for (sx_offset = 0; sx_offset < 8; sx_offset++) {
			b0 = *tileptr;
			b1 = *(tileptr + 8);

			if (hflip) {
				for (temp = 0; temp < sx_offset; temp++)
					b0 >>= 1, b1 >>= 1;
			} else {
				for (temp = 0; temp < (7 - sx_offset);
				     temp++)
					b0 >>= 1, b1 >>= 1;
			}

			b0 &= 0x01;
			b1 &= 0x01;
			ccolor = (c23 << 2) | (b1 << 1) | b0;

			if ((b1 | b0) == 0)
				ccolor = 0;
			else
				ccolor = nes_vram[0x3f10 + ccolor];

			if (ccolor != 0 && (sx + sx_offset) < 256) {
				dsprites[sx + sx_offset] = ccolor;
				dprior[sx + sx_offset] = priority;
			}
		}
	}

	return 0;
}


inline u8 nes_get_color_pixel_line(u8 y)
{
	register u8 *nt_ptr, *at_ptr, *td_ptr, *palette, *td_ptr_tmp;
	u16 nt_addr;
	u8 bg_color, ccolor;
	u8 tileno_x, tileno_y, tileno;
	u8 tileoff_x, tileoff_y;

	u8 a1, b1, b2, c1, c2, c3;
	u8 j, i, val;

	register u32 x;
	register u32 virt_x = BGSCROLLX;
	u16 virt_y = y + BGSCROLLY;

	u8 valyshift;
	u16 valyshift2;

	nt_addr = ((PPUCNT1 & 0x03) << 10) | 0x2000;
	palette = &nes_vram[0x3F00];

	td_ptr = &nes_vram[(PPUCNT1 & 0x10) << 8];

	bg_color = palette[0];

	if (virt_y > 239) {
		virt_y -= 240;
		nt_addr ^= 0x0800;
	}
	tileno_y = virt_y >> 3;
	tileoff_y = virt_y & 7;
	td_ptr += tileoff_y;	/* Account for Y-offset */


	c2 = tileno_y & 0x02;

	valyshift = ((tileno_y >> 2) << 3);
	valyshift2 = (tileno_y << 5);

	nt_ptr = nes_vram + nt_addr;
	at_ptr = nt_ptr + 0x03C0 + valyshift;

	nt_ptr += valyshift2;

	for (x = 0; x < 256; x++, virt_x++) {
		if (dsprites[x]) {
			dbackground[x] = dsprites[x];
			continue;
		}

		if (virt_x & 0xFF00) {
			nt_addr ^= 0x0400;
			virt_x &= 0xFF;
			nt_ptr = nes_vram + nt_addr;
			at_ptr = nt_ptr + 0x03C0 + valyshift;
			nt_ptr += valyshift2;
		}


		tileoff_x = virt_x & 7;


		tileno_x = virt_x >> 3;
		tileno = nt_ptr[tileno_x];
		td_ptr_tmp = td_ptr + (tileno << 4);

		b1 = (*td_ptr_tmp);
		b2 = (*(td_ptr_tmp + 8));

		b1 >>= (7 - tileoff_x);
		b2 >>= (7 - tileoff_x);
		b1 &= 1;
		b2 &= 1;

		a1 = at_ptr[(tileno_x >> 2)];
		c1 = (tileno_x & 0x02) >> 1;
		c3 = c2 | c1;
		val = c3 << 1;
		a1 = (a1 >> val) & 0x03;

		ccolor = (a1 << 2) | (b2 << 1) | b1;

		if (b1 | b2) {
			dbackground[x] = palette[ccolor];
		} else
			dbackground[x] = bg_color;

	}
	return 0;
}


inline void nes_bg_line(u8 line)
{
	register u8 *nt_ptr, *at_ptr, *td_ptr, *palette, *td_ptr_tmp;
	u16 nt_addr;
	u8 bg_color, ccolor;
	u8 tileno_x, tileno_y, tileno;
	u8 tileoff_x, tileoff_y;

	u8 a1, b1, b2, c1, c2, c3, z1 = 0, z2 = 0;
	u8 j, i, val;

	register u32 x = 0;
#ifdef DONTSCROLL
	register u32 virt_x = 0;
	u16 virt_y = line;
#else
	register u32 virt_x = BGSCROLLX & 0xF8;
	u16 virt_y = line + BGSCROLLY;
#endif
	u8 valyshift;
	u16 valyshift2;

	nt_addr = ((PPUCNT1 & 0x03) << 10) | 0x2000;
	if (nes_single_screen_mirroring == 1) {
		nt_addr = 0x2000;
	} else if (nes_single_screen_mirroring == 2) {
		nt_addr = 0x2400;
	}

	palette = &nes_vram[0x3F00];

	td_ptr = &nes_vram[(PPUCNT1 & 0x10) << 8];

	bg_color = palette[0];

	if (virt_y > 239) {
		virt_y -= 240;
		nt_addr ^= 0x0800;
	}
	tileno_y = virt_y >> 3;
	tileoff_y = virt_y & 7;
	td_ptr += tileoff_y;	/* Account for Y-offset */


	c2 = tileno_y & 0x02;

	valyshift = ((tileno_y >> 2) << 3);
	valyshift2 = (tileno_y << 5);

	nt_ptr = nes_vram + nt_addr;
	at_ptr = nt_ptr + 0x03C0 + valyshift;

	nt_ptr += valyshift2;

#ifndef DONTSCROLL
	if (BGSCROLLX & 0x07) {
		tileoff_x = BGSCROLLX & 0x07;

		tileno_x = virt_x >> 3;
		tileno = nt_ptr[tileno_x];
		td_ptr_tmp = td_ptr + (tileno << 4);

		b1 = (*td_ptr_tmp);
		b2 = (*(td_ptr_tmp + 8));

		a1 = at_ptr[(tileno_x >> 2)];
		c1 = (tileno_x & 0x02) >> 1;
		c3 = c2 | c1;
		val = c3 << 1;
		a1 = (a1 >> val) & 0x03;

		for (i = 0; i < 7 - (BGSCROLLX & 0x07);
		     i++, x++, tileoff_x++) {
			if (dsprites[x]) {
				dbackground[x] = dsprites[x];
				continue;
			}

			z1 = (b1 >> (7 - tileoff_x)) & 1;
			z2 = (b2 >> (7 - tileoff_x)) & 1;

			ccolor = (a1 << 2) | (z2 << 1) | z1;

			if (z1 | z2) {
				dbackground[x] = palette[ccolor];
			} else
				dbackground[x] = bg_color;
		}

		x = 7 - (BGSCROLLX & 0x07);
		virt_x += 8;
	} else
#endif
		x = 0;

	for (; x < 256;) {
		if (virt_x & 0xFF00) {
			nt_addr ^= 0x0400;
			virt_x &= 0xFF;
			nt_ptr = nes_vram + nt_addr;
			at_ptr = nt_ptr + 0x03C0 + valyshift;
			nt_ptr += valyshift2;
		}


		tileoff_x = 0;


		tileno_x = virt_x >> 3;
		tileno = nt_ptr[tileno_x];
		td_ptr_tmp = td_ptr + (tileno << 4);

		b1 = (*td_ptr_tmp);
		b2 = (*(td_ptr_tmp + 8));

		a1 = at_ptr[(tileno_x >> 2)];
		c1 = (tileno_x & 0x02) >> 1;
		c3 = c2 | c1;
		val = c3 << 1;
		a1 = (a1 >> val) & 0x03;

		for (i = 0; i < 8; i++, x++, virt_x++, tileoff_x++) {
			if (dsprites[x] && !dprior[x]) {
				dbackground[x] = dsprites[x];
				continue;
			}

			z1 = (b1 >> (7 - tileoff_x)) & 1;
			z2 = (b2 >> (7 - tileoff_x)) & 1;

			ccolor = (a1 << 2) | (z2 << 1) | z1;

			if (z1 | z2) {
				dbackground[x] = palette[ccolor];
			} else
				dbackground[x] = bg_color;

			if (dsprites[x] && dprior[x] && (z2 | z1) == 0)
				dbackground[x] = dsprites[x];
		}

	}
}
