#ifndef TARGET_DC

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>


#include "types.h"
#include "gleam!.h"
#include "mem.h"
#include "emulate.h"
#include "soundlin.h"

#include "oss.c"

int sq[4][8] = {
	{1, 0, 0, 0, 0, 0, 0, 0},
	{1, 1, 0, 0, 0, 0, 0, 0},
	{1, 1, 1, 1, 0, 0, 0, 0},
	{1, 1, 1, 1, 1, 1, 0, 0}
};

int lentab[32] = {
	0x05, 0x7F, 0x0A, 0x01, 0x14, 0x02, 0x28, 0x03,
	0x50, 0x04, 0x1E, 0x05, 0x07, 0x06, 0x0E, 0x07,
	0x06, 0x08, 0x0C, 0x09, 0x18, 0x0A, 0x30, 0x0B,
	0x60, 0x0C, 0x24, 0x0D, 0x08, 0x0E, 0x10, 0x0F
};

int voltab[16] =
    { 0, 0x0808, 0x1111, 0x1919, 0x2222, 0x2A2A, 0x3333, 0x3B3B,
	0x4444, 0x4C4C, 0x5555, 0x5D5D, 0x6666, 0x6E6E, 0x7777, 0x7F7F
};

int noitab[16] = { 0x002, 0x004, 0x008, 0x010, 0x020, 0x030, 0x040, 0x050,
	0x065, 0x07F, 0x0BE, 0x0FE, 0x17D, 0x1FC, 0x3F9, 0x7F2
};

int dmctab[16] = { 0xD60, 0xBE0, 0xAA0, 0xA00, 0x8F0, 0x7F0, 0x710, 0x6B0,
	0x5F0, 0x500, 0x470, 0x400, 0x350, 0x2A0, 0x240, 0x1B0
};

int scancyc = 115 * 65536;
int soundld = 2659739;		/* 1789772*65536/44100 */
int soundfrq = 1789772 / 44100;
int lenfrq = 1789772 / 60;
int envfrq = 1789772 / 240;
int swpfrq = 1789772 / 120;

int soundcnt;

int sq1_en, sq1_duty, sq1_len, sq1_lend, sq1_lcyc, sq1_freq, sq1_fcyc;
int sq1_vol, sq1_decay, sq1_envon, sq1_envrate, sq1_envloop, sq1_ecyc;
int sq1_swon, sq1_swshift, sq1_swdir, sq1_swrate, sq1_scyc;

int sq2_en, sq2_duty, sq2_len, sq2_lend, sq2_lcyc, sq2_freq, sq2_fcyc;
int sq2_vol, sq2_decay, sq2_envon, sq2_envrate, sq2_envloop, sq2_ecyc;
int sq2_swon, sq2_swshift, sq2_swdir, sq2_swrate, sq2_scyc;

int tri_en, tri_len, tri_lend, tri_lcyc, tri_freq, tri_fcyc;
int tri_linl, tri_lincyc, tri_linc, tri_linmode, tri_linload;

int noi_en, noi_len, noi_lend, noi_lcyc, noi_rate, noi_fcyc, noi_type;
int noi_vol, noi_decay, noi_envon, noi_envrate, noi_envloop, noi_ecyc;
int noi_sr = 0x0001, noi_bit;

int dmc_en, dmc_mode, dmc_freq, dmc_fcyc, dmc_val, dmc_addr, dmc_addrs;
int dmc_len, dmc_lens, dmc_lcyc, dmc_irq, dmc_byte, dmc_mask;

void gleam_sound_write(u8 reg, u8 data)
{
	switch (reg) {
	case 0x00:
		sq1_lend = (data & 0x20) ? 1 : 0;
		sq1_duty = data >> 6;
		sq1_envloop = (data & 0x20) ? 1 : 0;
		sq1_envon = (data & 0x10) ? 0 : 1;
		if (sq1_envon) {
			sq1_envrate = data & 0x0F;
		} else {
			sq1_vol = data & 0x0F;
		}
		sq1_ecyc = 0;
		break;
	case 0x01:
		sq1_swshift = data & 0x03;
		sq1_swdir = (data & 0x08) ? 1 : 0;
		sq1_swrate = (data & 0x70) >> 4;
		sq1_swon = (data & 0x80) ? 1 : 0;
		sq1_scyc = 0;
		break;
	case 0x02:
		sq1_freq = (sq1_freq & 0x700) | data;
		//sq1_fcyc = 0;
		break;
	case 0x03:
		sq1_freq = (sq1_freq & 0x0FF) | ((data & 0x07) << 8);
		//sq1_fcyc = 0;
		sq1_len = lentab[data >> 3];
		sq1_lcyc = 0;
		sq1_decay = 0x0F;
		sq1_ecyc = 0;
		break;
	case 0x04:
		sq2_lend = (data & 0x20) ? 1 : 0;
		sq2_duty = data >> 6;
		sq2_envloop = (data & 0x20) ? 1 : 0;
		sq2_envon = (data & 0x10) ? 0 : 1;
		if (sq2_envon) {
			sq2_envrate = data & 0x0F;
		} else {
			sq2_vol = data & 0x0F;
		}
		sq2_ecyc = 0;
		break;
	case 0x05:
		sq2_swshift = data & 0x03;
		sq2_swdir = (data & 0x08) ? 1 : 0;
		sq2_swrate = (data & 0x70) >> 4;
		sq2_swon = (data & 0x80) ? 1 : 0;
		sq2_scyc = 0;
		break;
	case 0x06:
		sq2_freq = (sq2_freq & 0x700) | data;
		//sq2_fcyc = 0;
		break;
	case 0x07:
		sq2_freq = (sq2_freq & 0x0FF) | ((data & 0x07) << 8);
		//sq2_fcyc = 0;
		sq2_len = lentab[data >> 3];
		sq2_lcyc = 0;
		sq2_decay = 0x0F;
		sq2_ecyc = 0;
		break;
	case 0x08:
		tri_linload = data & 0x7F;
		if (!tri_lend)
			tri_linmode = 1;
		else if (data & 0x80)
			tri_linmode = 1;
		tri_lend = (data & 0x80) ? 1 : 0;
		tri_lincyc = 0;
		break;
	case 0x09:
		break;
	case 0x0A:
		tri_freq = (tri_freq & 0x700) | data;
		break;
	case 0x0B:
		tri_freq = (tri_freq & 0x0FF) | ((data & 0x07) << 8);
		tri_len = lentab[data >> 3];
		tri_linl = tri_linload;
		if (tri_lend)
			tri_linmode = 0;
		else
			tri_linmode = 1;
		tri_lcyc = 0;
		tri_lincyc = 0;
		break;
	case 0x0C:
		noi_lend = (data & 0x20) ? 1 : 0;
		noi_envloop = (data & 0x20) ? 1 : 0;
		noi_envon = (data & 0x10) ? 0 : 1;
		if (noi_envon) {
			noi_envrate = data & 0x0F;
		} else {
			noi_vol = data & 0x0F;
		}
		noi_ecyc = 0;
		break;
	case 0x0D:
		break;
	case 0x0E:
		noi_rate = data & 0x0F;
		noi_type = (data & 0x80) ? 1 : 0;
		break;
	case 0x0F:
		noi_len = lentab[data >> 3];
		noi_decay = 0x0F;
		break;
	case 0x10:
		printf("data is %02X\n", data);
		dmc_mode = data >> 6;
		dmc_freq = data & 0x0F;
		break;
	case 0x11:
		dmc_val = data & 0x7F;
		break;
	case 0x12:
		dmc_addrs = (data << 6) | 0xC000;
		break;
	case 0x13:
		dmc_lens = (data << 4) + 1;
		break;
	case 0x15:
		if (data & 0x01) {
			sq1_en = 1;
		} else {
			sq1_en = 0;
			sq1_len = 0;
		}
		if (data & 0x02) {
			sq2_en = 1;
		} else {
			sq2_en = 0;
			sq1_len = 0;
		}
		if (data & 0x04) {
			tri_en = 1;
		} else {
			tri_en = 0;
			tri_len = 0;
		}
		if (data & 0x08) {
			noi_en = 1;
		} else {
			noi_en = 0;
			noi_len = 0;
		}
		if (data & 0x10) {
			dmc_en = 1;
		} else {
			dmc_en = 0;
			dmc_len = 0;
		}
		break;
	default:
		printf("Unknown sound write: ");
		printf("SOUND: Wrote %02X to Sound Register 0x40%02X\n",
		       data, reg);
		break;
	}

	nes_soundregs[reg] = data;
}

u8 gleam_sound_read(u8 reg)
{
	fprintf(stdout, "SOUND: Read Sound Register 0x40%02X\n", reg);
	if (reg == 0x15) {
		return
		    (sq1_len ? 0x01 : 0x00) |
		    (sq2_len ? 0x02 : 0x00) |
		    (tri_len ? 0x04 : 0x00) |
		    (noi_len ? 0x08 : 0x00) |
		    (dmc_len ? 0x10 : 0x00) | (dmc_irq ? 0x80 : 0x00);
	}
	return 0;
}

/*
Called every scanline,
*/

void WriteSample(short int l, short int r);

void gleam_process_sound()
{
	int bit;
	int l, r, m;
	int x;

	soundcnt += scancyc;

	while (soundcnt >= soundld) {
		l = 0;
		r = 0;
#if 1
		if (sq1_en && (sq1_lend || sq1_len)) {
			if (sq1_freq > 3) {
				bit =
				    sq[sq1_duty][((sq1_fcyc /
						   (sq1_freq +
						    1)) >> 2) & 0x07];
				x = sq1_envon ? sq1_decay : sq1_vol;
				if (bit)
					l += voltab[x];
				else
					l -= voltab[x];
			}

			sq1_fcyc += soundfrq;
			while (sq1_fcyc >= 32 * (sq1_freq + 1)) {
				sq1_fcyc -= 32 * (sq1_freq + 1);
			}

			if (!sq1_lend) {
				sq1_lcyc += soundfrq;
				while (sq1_lcyc >= lenfrq) {
					sq1_lcyc -= lenfrq;
					sq1_len--;
					if (sq1_len == 0)
						break;
				}
			}

			if ( /*sq1_envon && */ (sq1_decay || sq1_envloop)) {
				sq1_ecyc += soundfrq;
				while (sq1_ecyc >=
				       (sq1_envrate + 1) * envfrq) {
					sq1_ecyc -=
					    (sq1_envrate + 1) * envfrq;
					sq1_decay--;
					if (sq1_decay == 0)
						if (!sq1_envloop)
							break;
					if (sq1_decay == -1)
						sq1_decay = 0x0F;
				}
			}

			if (sq1_swon && sq1_swshift) {
				sq1_scyc += soundfrq;
				while (sq1_scyc >=
				       (sq1_swrate + 1) * swpfrq) {
					sq1_scyc -=
					    (sq1_swrate + 1) * swpfrq;
					if (sq1_swdir)
						sq1_freq -=
						    ((sq1_freq >>
						      sq1_swshift) + 1);
					else
						sq1_freq +=
						    (sq1_freq >>
						     sq1_swshift);
					if (sq1_freq < 8) {
						sq1_en = 0;
						break;
					}
					if (sq1_freq > 2047) {
						sq1_en = 0;
						break;
					}
				}
			}
		}
#endif

#if 1
		if (sq2_en && (sq2_lend || sq2_len)) {
			if (sq2_freq > 3) {
				bit =
				    sq[sq2_duty][((sq2_fcyc /
						   (sq2_freq +
						    1)) >> 2) & 0x07];
				x = sq2_envon ? sq2_decay : sq2_vol;
				if (bit)
					l += voltab[x];
				else
					l -= voltab[x];
			}

			sq2_fcyc += soundfrq;
			while (sq2_fcyc >= 32 * (sq2_freq + 1)) {
				sq2_fcyc -= 32 * (sq2_freq + 1);
			}

			if (!sq2_lend) {
				sq2_lcyc += soundfrq;
				while (sq2_lcyc >= lenfrq) {
					sq2_lcyc -= lenfrq;
					sq2_len--;
					if (sq2_len == 0)
						break;
				}
			}

			if ( /*sq2_envon && */ (sq2_decay || sq2_envloop)) {
				sq2_ecyc += soundfrq;
				while (sq2_ecyc >=
				       (sq2_envrate + 1) * envfrq) {
					sq2_ecyc -=
					    (sq2_envrate + 1) * envfrq;
					sq2_decay--;
					if (sq2_decay == 0)
						if (!sq2_envloop)
							break;
					if (sq2_decay == -1)
						sq2_decay = 0x0F;
				}
			}

			if (sq2_swon && sq2_swshift) {
				sq2_scyc += soundfrq;
				while (sq2_scyc >=
				       (sq2_swrate + 1) * swpfrq) {
					sq2_scyc -=
					    (sq2_swrate + 1) * swpfrq;
					if (sq2_swdir)
						sq2_freq -=
						    ((sq2_freq >>
						      sq2_swshift) + 1);
					else
						sq2_freq +=
						    (sq2_freq >>
						     sq2_swshift);
					if (sq2_freq < 8) {
						sq2_en = 0;
						break;
					}
					if (sq2_freq > 2047) {
						sq2_en = 0;
						break;
					}
				}
			}
		}
#endif

#if 1
		if (tri_en && ((tri_lend && tri_linl) || tri_len)) {
			if (tri_freq > 4) {
				if (tri_linc & 0x10) {
					x = (~tri_linc) & 0x0F;
				} else {
					x = tri_linc & 0x0F;
				}

				r += voltab[x] - 16384;
			}

			tri_fcyc += soundfrq;
			while (tri_fcyc >= (tri_freq + 1)) {
				tri_fcyc -= (tri_freq + 1);
				tri_linc++;
			}
			tri_linc &= 0x1F;

			if (!tri_lend) {
				tri_lcyc += soundfrq;
				while (tri_lcyc >= lenfrq) {
					tri_lcyc -= lenfrq;
					tri_len--;
					if (tri_len == 0)
						break;
				}
			}
			if (tri_linmode) {
				tri_lincyc += soundfrq;
				while (tri_lincyc >= lenfrq) {
					tri_lincyc -= lenfrq;
					tri_linl--;
					if (tri_linl == 0)
						break;
				}
			}

		}
#endif

#if 1
		if (noi_en && (noi_lend || noi_len)) {
			x = noi_envon ? noi_decay : noi_vol;

			if (noi_bit)
				r += voltab[x];
			else
				r -= voltab[x];

			noi_fcyc += soundfrq;
			while (noi_fcyc >= 2 * noitab[noi_rate]) {
				int bit1, bit2;

				noi_fcyc -= 2 * noitab[noi_rate];

				noi_bit = (noi_sr & 0x4000) ? 0 : 1;
				noi_sr <<= 1;
				if (noi_type) {
					bit1 = (noi_sr & 0x8000) ? 1 : 0;
					bit2 = (noi_sr & 0x0200) ? 1 : 0;
				} else {
					bit1 = (noi_sr & 0x8000) ? 1 : 0;
					bit2 = (noi_sr & 0x4000) ? 1 : 0;
				}
				noi_sr &= 0x7FFF;
				noi_sr |= (bit1 ^ bit2);
			}

			if (!noi_lend) {
				noi_lcyc += soundfrq;
				while (noi_lcyc >= lenfrq) {
					noi_lcyc -= lenfrq;
					noi_len--;
					if (noi_len == 0)
						break;
				}
			}

			if ( /*noi_envon && */ (noi_decay || noi_envloop)) {
				noi_ecyc += soundfrq;
				while (noi_ecyc >=
				       (noi_envrate + 1) * envfrq) {
					noi_ecyc -=
					    (noi_envrate + 1) * envfrq;
					noi_decay--;
					if (noi_decay == 0)
						if (!noi_envloop)
							break;
					if (noi_decay == -1)
						noi_decay = 0x0F;
				}
			}
		}
#endif

#if 0
		if (dmc_en && dmc_len) {
			if (!dmc_mask) {
				dmc_byte = nes_read_byte(dmc_addr);
				dmc_addr++;
				dmc_addr &= 0xFFFF;
				dmc_addr |= 0x8000;
				dmc_mask = 0x01;
				dmc_len--;
			}

			r += (dmc_data << 8) - 16384;

			dmc_fcyc += soundfrq;
			while (dmc_fcyc >= (dmctab[dmc_freq] / 8)) {
				dmc_fcyc -= (dmctab[dmc_freq] / 8);
				if (dmc_mask & dmc_byte)
					dmc_data += 2;
				else
					dmc_data -= 2;
				if ((dmc_data & ~1) > 0x7E)
					dmc_data &= 0x7F;
				else if ((dmc_data & ~1) < 0)
					dmc_data &= 0x01;
				dmc_mask <<= 1;
			}
		}
#endif
		m = l + r;
		m >>= 3;

		l >>= 2;
		r >>= 2;
		l += m;
		r += m;

		soundcnt -= soundld;

		WriteSample(l, r);
	}
}


#endif				/* TARGET_DC */
