#ifndef TARGET_DC

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/extensions/XShm.h>

#include "gleam!.h"

#include "emulate.h"

Display *dpy;
int screen;
Window root, win;
XVisualInfo *vi;
GC gc;
XImage *img;
Colormap cmap;
Atom wmDeleteWindow;
XShmSegmentInfo shminfo;
XColor clr[256];

short int spal[256];

short int *sgfx;

unsigned long option_frameskip = 1;

static unsigned long
 gleam_window_width = 256, gleam_window_height = 240;

int gleam_init_gl(char *display)
{
	dpy = XOpenDisplay(display);

	if (dpy == NULL) {
		fprintf(stderr, "Failed to open X connect to %s\n",
			XDisplayName(display));
		return 0;
	}

	screen = DefaultScreen(dpy);
	root = RootWindow(dpy, screen);

	return 1;
}

int gleam_close_gl()
{
	if (shminfo.shmaddr) {
		XShmDetach(dpy, &shminfo);
		shmdt(shminfo.shmaddr);
		shmctl(shminfo.shmid, IPC_RMID, 0);
	}
	shminfo.shmaddr = NULL;

	if (img) {
		XDestroyImage(img);
	}
	img = NULL;

	return 0;
}

void GetVisual()
{
	XVisualInfo vitemp;
	int numVisuals, mask, cmask, i;

	mask = VisualScreenMask | VisualDepthMask | VisualClassMask;
	cmask =
	    VisualRedMaskMask | VisualGreenMaskMask | VisualBlueMaskMask;

	vitemp.screen = screen;
	vitemp.depth = 4;
	vitemp.class = PseudoColor;

	vi = XGetVisualInfo(dpy, mask, &vitemp, &numVisuals);
	if (vi && (numVisuals > 0)) {
		cmap = XCreateColormap(dpy, root, vi->visual, AllocAll);

		for (i = 0; i < 16; i++) {
			clr[i].pixel = i;
			clr[i].flags = DoRed | DoGreen | DoBlue;
		}
		return;
	}

	vitemp.depth = 8;

	vi = XGetVisualInfo(dpy, mask, &vitemp, &numVisuals);
	if (vi && (numVisuals > 0)) {
		cmap = XCreateColormap(dpy, root, vi->visual, AllocAll);

		for (i = 0; i < 256; i++) {
			clr[i].pixel = i;
			clr[i].flags = DoRed | DoGreen | DoBlue;
		}
		return;
	}

	vitemp.depth = 15;
	vitemp.class = TrueColor;
	vitemp.red_mask = 0x7C00;
	vitemp.green_mask = 0x03E0;
	vitemp.blue_mask = 0x001F;

	vi = XGetVisualInfo(dpy, mask | cmask, &vitemp, &numVisuals);
	if (vi && (numVisuals > 0)) {
		cmap = XCreateColormap(dpy, root, vi->visual, AllocNone);
		return;
	}

	vitemp.depth = 16;
	vitemp.red_mask = 0xF800;
	vitemp.green_mask = 0x07E0;
	vitemp.blue_mask = 0x001F;

	vi = XGetVisualInfo(dpy, mask | cmask, &vitemp, &numVisuals);
	if (vi && (numVisuals > 0)) {
		cmap = XCreateColormap(dpy, root, vi->visual, AllocNone);
		return;
	}

	vitemp.depth = 24;
	vitemp.red_mask = 0xFF0000;
	vitemp.green_mask = 0x00FF00;
	vitemp.blue_mask = 0x0000FF;

	vi = XGetVisualInfo(dpy, mask | cmask, &vitemp, &numVisuals);
	if (vi && (numVisuals > 0)) {
		cmap = XCreateColormap(dpy, root, vi->visual, AllocNone);
		return;
	}

	vitemp.depth = 32;

	vi = XGetVisualInfo(dpy, mask | cmask, &vitemp, &numVisuals);
	if (vi && (numVisuals > 0)) {
		cmap = XCreateColormap(dpy, root, vi->visual, AllocNone);
		return;
	}

	fprintf(stderr, "No usable visuals found!\n");
	exit(EXIT_FAILURE);
}

int gleam_create_window_gl(char *title, int width, int height)
{
	XSetWindowAttributes attr;
	XSizeHints sizehints;
	XGCValues gcvalues;
	Pixmap bitmap;
	Cursor cursor;
	XColor bg = { 0 };
	XColor fg = { 0 };
	char data[8] = { 0x01 };
	int mask;

	GetVisual();

	if (vi->depth != 16) {
		fprintf(stderr, "Only 16bpp currently supported!\n");
		exit(EXIT_FAILURE);
	}

	attr.colormap = cmap;
	attr.event_mask =
	    KeyPressMask | KeyReleaseMask | ExposureMask |
	    StructureNotifyMask;
	mask = CWEventMask | CWColormap;

	win = XCreateWindow(dpy, root, 0, 0, width, height, 0, vi->depth,
			    InputOutput, vi->visual, mask, &attr);

	if (win == None) {
		printf("Unable to create X window!\n");
		return 0;
	}

	sizehints.min_width = width;
	sizehints.min_height = height;
	sizehints.max_width = width;
	sizehints.max_height = height;
	sizehints.base_width = width;
	sizehints.base_height = height;
	sizehints.flags = PMinSize | PMaxSize | PBaseSize;

	XSetWMProperties(dpy, win, NULL, NULL, NULL, 0, &sizehints, None,
			 None);

	XStoreName(dpy, win, title);
	XSetIconName(dpy, win, title);

	wmDeleteWindow = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
	XSetWMProtocols(dpy, win, &wmDeleteWindow, 1);

	gcvalues.foreground = BlackPixel(dpy, screen);
	gcvalues.background = WhitePixel(dpy, screen);
	gc = XCreateGC(dpy, win, GCForeground | GCBackground, &gcvalues);

	bitmap = XCreateBitmapFromData(dpy, win, data, 8, 8);
	cursor = XCreatePixmapCursor(dpy, bitmap, bitmap, &fg, &bg, 0, 0);
	XDefineCursor(dpy, win, cursor);

	XMapRaised(dpy, win);

	XFlush(dpy);

	sgfx = (short int *) malloc(width * height * 2);
	img =
	    XCreateImage(dpy, vi->visual, vi->depth, ZPixmap, 0,
			 (char *) sgfx, width, height, 8, width * 2);

	reset_colors();

	return 1;
}

static void handler_keypress(KeySym key)
{
	extern float vblmax;
	switch (key) {
	case XK_Escape:
		printf("\nmax fps = %f\n", vblmax);
		exit(1);
		break;
	case XK_Return:
		JOYPAD1 |= 0x08;
		break;
	case XK_Tab:
		JOYPAD1 |= 0x04;
		break;
	case XK_a:
	case XK_Control_L:
		JOYPAD1 |= 0x02;
		break;
	case XK_f:
	case XK_Alt_L:
		JOYPAD1 |= 0x01;
		break;
	case XK_Left:
		JOYPAD1 |= 0x40;
		break;
	case XK_Right:
		JOYPAD1 |= 0x80;
		break;
	case XK_Up:
		JOYPAD1 |= 0x10;
		break;
	case XK_Down:
		JOYPAD1 |= 0x20;
		break;
	}
}

static void handler_keyunpress(KeySym key)
{
	switch (key) {
	case XK_Return:
		JOYPAD1 &= ~0x08;
		break;
	case XK_Tab:
		JOYPAD1 &= ~0x04;
		break;
	case XK_a:
	case XK_Control_L:
		JOYPAD1 &= ~0x02;
		break;
	case XK_f:
	case XK_Alt_L:
		JOYPAD1 &= ~0x01;
		break;
	case XK_Left:
		JOYPAD1 &= ~0x40;
		break;
	case XK_Right:
		JOYPAD1 &= ~0x80;
		break;
	case XK_Up:
		JOYPAD1 &= ~0x10;
		break;
	case XK_Down:
		JOYPAD1 &= ~0x20;
		break;
	}
}

int gleam_poll_events()
{
	XEvent event;

	while (XPending(dpy)) {
		XNextEvent(dpy, &event);

		/* TODO: watch focus and clear key status when focus changes */
		switch (event.type) {
		case KeyPress:
			handler_keypress(XKeycodeToKeysym
					 (dpy, event.xkey.keycode, 0));
			break;
		case KeyRelease:
			handler_keyunpress(XKeycodeToKeysym
					   (dpy, event.xkey.keycode, 0));
			break;
		case ConfigureNotify:
			break;
		case ClientMessage:
			if (event.xclient.data.l[0] == wmDeleteWindow)
				handler_keypress(XK_Escape);	/* HACK */
			break;
		default:
			/*printf("What is this: %d\n", event.type); */
			break;
		}
		XFlush(dpy);
	}

	return 1;
}

void vram_blit()
{
	XPutImage(dpy, win, gc, img, 0, 0, 0, 0, gleam_window_width,
		  gleam_window_height);
}

void set_color(int num, unsigned short int red, unsigned short int green,
	       unsigned short int blue)
{
	red >>= 1;
	blue >>= 1;
	spal[num] = (red << 11) | (green << 5) | blue;
}

void vram_plot_line(int y, u8 * color)
{
	short int *sptr = &sgfx[y * gleam_window_width];
	int i;

	for (i = 0; i < 256; i++)
		sptr[i] = spal[color[i]];
}

void vram_vline(short int *ptr, int color, int len)
{
	int i;
	for (i = 0; i < len; i++)
		ptr[i] = spal[color];
}

void vram_hline(short int *ptr, int color, int height)
{
	int i;
	for (i = 0; i < height; i++) {
		ptr[i * gleam_window_width] = spal[color];
	}
}

void vram_box(int x, int y)
{
	short int *pstr;
	int i;

	pstr = &sgfx[x + (y * gleam_window_width)];

	for (i = 0; i < 256; i++)
		pstr[i] = spal[0];
	for (i = 0; i < 240; i++)
		pstr[(i * gleam_window_width)] =
		    pstr[255 + (i * gleam_window_width)] = spal[0];
	for (i = 0; i < 256; i++)
		pstr[(239 * gleam_window_width) + i] = spal[0];
}

#endif				/* TARGET_DC */
