/* This file is part of the Dreamcast function library.
 * Please see libdream.c for further details.
 *
 * (c)2000 Dan Potter
 */

#ifndef __DREAM_H
#define __DREAM_H

/*********** General setup *****/
#ifndef NULL
#define NULL (void *)0
#endif

typedef unsigned long uint32;
typedef unsigned short uint16;
typedef unsigned char uint8;
typedef signed long int32;
typedef signed short int16;
typedef signed char int8;

/*********** libdream.c ********/
void dc_setup(int video_mode);
void dc_sleep(int seconds);

/*********** video.c ***********/

extern unsigned long *vram_l;
extern unsigned short *vram_s;
extern int vram_config;
extern int vram_size;

#define CT_VGA		0
#define CT_RGB		2
#define CT_COMPOSITE	3
int dc_vid_check_cable();

#define PM_RGB555	0
#define PM_RGB565	1
#define PM_RGB888	3
void dc_vid_init(int cable_type, int pixel_mode);

void dc_vid_border_color(int r, int g, int b);
void dc_vid_clear(int r, int g, int b);
void dc_vid_waitvbl();

/************ tty.c ************/
void dc_serial_init();
void dc_serial_printf(char *fmt, ...);


/************ font.c ***********/
void dc_font_set(unsigned char * fbm, int fh);
void dc_draw_char(int x1, int y1, unsigned short color, int ch, unsigned short *);
void dc_draw_string(int x1, int y1, unsigned short color, char *str, unsigned short *);
void dc_draw_stringf(unsigned short *, int x1, int y1, unsigned short color, char *fmt, ...);

/************ spu.c ************/
#define SM_16BIT	0		/* signed 16-bit words */
#define SM_8BIT		1		/* signed 8-bit bytes */
#define SM_ADPCM	3		/* ADPCM compressed samples */
#define SM_STATIC	(0xff << 1)	/* Static test pattern; can be used
					   for a really simple chip-player =) */
#define SMP_BASE	0x10000

void dc_snd_base_arm();
void dc_snd_load_arm(void* src, int size);
void dc_snd_stop_arm();
void dc_snd_init();
void dc_snd_play(int ch, int smpbase, int mode, int nsamp, int freq, int vol, int pan);
void dc_snd_stop(int ch);
void dc_snd_vol(int ch, int vol);
void dc_snd_pan(int ch, int pan);
void dc_snd_freq(int ch, int freq);
void dc_snd_load(void *src, int dest, int len);

/************ cdfs.c ************/
#include "cdfs.h"

/************ maple.c ***********/
#include "maple.h"

/************ controller.c ******/
#include "controller.h"

/************ input.c ***********/
void dc_input_init();
void dc_input_scan();

#endif	// __DREAM_H




