/* This file is part of the Dreamcast function library.
 * Please see libdream.c for further details.
 *
 * (c)2000 Dan Potter
 */

#include "dream.h"

/* An array of all the attached devices; each element contains
   a function code or zero if nothing is attached. Note that if you
   swap around controllers after init, you'll have to call 
   input_scan() again. */
static int func_codes[4][6];

void dc_input_init() {
	dc_maple_init();
	dc_input_scan();
}


/* Scans one maple bus unit for its information. The values will
   be stored into the bus array.

   Info taken from Marcus' maple bus example */
void dc_input_scan_one(int port, int unit, char *res) {
	struct maple_devinfo *di = (struct maple_devinfo *)(res+4);
	int func = read_belong(&di->func) & MAPLE_FUNC_MASK;
	func_codes[port][unit] = func;
}

/* Scans the maple bus for peripherals and stores the info.

   Info taken from Marcus' maple bus example */
void dc_input_scan() {
	char *res;
	int port, unit;

	for (port=0; port<4; port++) {
		for (unit=0; unit<6; unit++) {
			do {
				res = dc_maple_docmd(port, unit,
					MAPLE_COMMAND_DEVINFO, 0, NULL);
			} while (*res == MAPLE_RESPONSE_AGAIN);

			if (*res == MAPLE_RESPONSE_DEVINFO && res[3] >= 28) {
				dc_input_scan_one(port, unit, res);
			}
		}
	}
}



/* A little convienence routine that waits for start on any controller.
   The controller (port id) that pressed start will be returned. */
int dc_input_wait_start() {
	int i;

	while(1) {
		for (i=0; i<4; i++) {
			short state = dc_cont_get_buttons(i);
			if (state == -1) continue;
			if (state & CONT_START)
				return i;
		}
	}
}

