/*
 * Dreamcast function library (libdream)
 * Portions (c)2000 Dan Potter, Jordan DeLong, and Mike Brent
 *
 * Many thanks to and helpful info from these guys:
 *  Marcus Comstedt (cdfs.c/h, maple.c/h, some sound and video info)
 *  Bitmaster (video info)
 *  HeroZero (waitvbl info)
 *  HitMen-DC (help on gcc)
 *  The DCDev list on eGroups 
 *
 * This library is ilcensed under the MIT/X style license. For more
 * information, please see COPYING. Note that the above copyright notice
 * applies only to parts of the library that were created by those people.
 * Several complete source files in this library were imported from other
 * sources with no explicit copyright notice. When this happened, a notation
 * is made at the top of the file. Please contact those authors for 
 * information on using their code in your programs. Thanks!
 */


/* This pretty much does everything you need to do -- it grabs
   the cable spec, sets up the video, and initializes the serial
   port (for debugging). Returns 0 on success. */
int dc_setup(int video_mode) {
	/* Initialize SCIF */
	dc_serial_init();
	dc_serial_printf("SCIF debug interface initialized\r\n");
	
	/* Init video */
	dc_vid_init(dc_vid_check_cable(), video_mode);
	dc_vid_clear(0,0,0);
	
	/* Init sound */
	dc_snd_init();
	
	/* Init GD-ROM */
	dc_gd_init();

	/* Init input devices (also: Maple Bus) */
	dc_input_init();

	return 0;
}


/* A simple delay routine based on PVR wait states. Not very accurate =) */
void dc_sleep(int millis) {
	int i, v, cnt;
	
	cnt = 0x58e * millis / 1000;
	for (i=0; i<cnt; i++)
		dc_snd_ram_write_wait();
}



