! This file is distributed with the libdream Dreamcast function
! library, but it is not written by the author(s) of libdream. This
! file was written by Marcus Comstedt, with tweaks by Mike Brent.


	! Dreamcast Serial example by marcus
	!
	! Set up the serial interface and
	! send some data
	!
	! To compile into SREC using GNU binutils:
	!
	! sh-elf-as -little -o serial.o serial.s
	! sh-elf-ld -EL --oformat srec -Ttext 0x8c010000 serial.o -o serial.srec


	.globl	_InitSerial
	.globl  _WriteSerial
	.globl	_ReadSerial
	
	.text

	! Functions start here
_InitSerial:
	! Save Registers
	mov.l	r14,@-r15
	mov.l	r4, @-r15
	sts.l	pr,	@-r15
	mov		r15, r14

	! Setup serial port
	! Initialize SCIF registers
	!
	! r4 = baudrate (must be >= 6092 bps)
	!	mov.l	baud_rate,r4
	! Get base address of SCIF
	mov	#-24,r2
	shll16	r2
	! disable interrupts, disable transmit/receive, use internal clock
	mov	#0,r0
	mov.w	r0,@(8,r2)
	! 8N1, use P clock
	mov.w	r0,@r2
	! Set baudrate, N = P/(32*B)-1
	div0u
	mov.l	Pphi_32rds,r0
	mov	r4,r1
	shlr	r1
	add	r1,r0
	mov	#0,r1
	.rep	32
	rotcl	r0
	div1	r4,r1
	.endr
	rotcl	r0
	add	#-1,r0
	mov.b	r0,@(4,r2)
	! reset FIFOs, enable hardware flow control
	add	#8,r2
	mov	#12,r0
	mov.w	r0,@(16,r2)
	mov	#8,r0
	mov.w	r0,@(16,r2)
	! disable manual pin control
	mov	#0,r0
	mov.w	r0,@(24,r2)
	! clear status
	mov.w	@(8,r2),r0
	mov	#0x60,r0
	mov.w	r0,@(8,r2)
	mov.w	@(28,r2),r0
	mov	#0,r0
	mov.w	r0,@(28,r2)
	! enable transmit/receive
	mov	#0x30,r0
	mov.w	r0,@r2

	! restore registers
	mov		r14, r15
	lds.l	@r15+, pr
	mov.l	@r15+, r4
	mov.l	@r15+, r14

	rts
	nop


	! Write text to the serial port ;)
_WriteSerial:
	
	! set up stack pointer (from GCC output)
	mov.l	r14,@-r15
	mov.l	r4, @-r15
	sts.l	pr,	@-r15
	mov		r15, r14

	! pointer in R4
	bsr	send_str
	nop

	! Make sure everything has actually been sent
	! to the port before returning

	bsr	flush
	nop

	! fix up stack for return
	mov		r14,	r15
	lds.l	@r15+,	pr
	mov.l	@r15+,	r4
	mov.l	@r15+,	r14

	rts	
	nop


	.align	4
baud_rate:	
	.long	57600

Pphi_32rds:
	.long	50000000/32


	! Send a NUL-terminated string to the serial port
	!
	! r4 = string
send_str:
	mov	#-24,r2
	shll16	r2
	add	#12,r2
.ssloop1:
	mov.b	@r4+,r1
	tst	r1,r1
	bt	.endstr
	extu.b	r1,r1
.ssloop2:
	mov.w	@(4,r2),r0
	tst	#0x20,r0
	bt	.ssloop2
	mov.b	r1,@r2
	nop
	and	#0x9f,r0
	bra	.ssloop1
	mov.w	r0,@(4,r2)
.endstr:
	nop
	rts
	nop


	! Send a single character to the serial port
	!
	! r4 = character
send:
	mov	#-24,r2
	shll16	r2
	add	#12,r2
.sloop:	
	mov.w	@(4,r2),r0
	tst	#0x20,r0
	bt	.sloop
	mov.b	r4,@r2
	nop
	and	#0x9f,r0
	rts
	mov.w	r0,@(4,r2)


	! Flush output FIFO
flush:
	mov	#-24,r2
	shll16	r2
	add	#12,r2
	mov.w	@(4,r2),r0
	and	#0xbf,r0
	mov.w	r0,@(4,r2)
.wflush:	
	mov.w	@(4,r2),r0
	tst	#64,r0
	bt	.wflush	
	and	#0xbf,r0
	rts
	mov.w	r0,@(4,r2)
	
	
	! Check for serial input
	!
	! returns character or -1
_ReadSerial:
recv:
	mov.l	r14,@-r15
	sts.l	pr, @-r15
	mov.l	r1,@-r15
	mov.l	r2,@-r15
	
	mov	#-24,r2
	shll16	r2
	add	#12,r2
	mov.w	@(16,r2),r0
	tst	#0x1f,r0
	bt	.norecv
	mov.b	@(8,r2),r0
	extu.b	r0,r1
	mov.w	@(4,r2),r0
	and	#0x6d,r0
	mov.w	r0,@(4,r2)
	mov	r1,r0
	
	mov.l	@r15+,r2
	mov.l	@r15+,r1
	lds.l	@r15+,pr
	rts
	mov.l	@r15+,r14
.norecv:
	mov	#-1,r0
	mov.l	@r15+,r2
	mov.l	@r15+,r1
	lds.l	@r15+,pr
	rts
	mov.l	@r15+,r14
	
	.end

	
